import struct

CLIENT_ID = 49151

# Protocol adapter closes the connection if it doesn't hear
# from the client for 10 minutes, but we'll go ahead an make
# sure it hears something from us well before then.
COM_KEEPALIVE = 300  # Seconds = 5 minutes

SL_GATEWAY_IP = "ip"
SL_GATEWAY_PORT = "port"
SL_GATEWAY_TYPE = "gtype"
SL_GATEWAY_SUBTYPE = "gsubtype"
SL_GATEWAY_NAME = "name"


class ScreenLogicError(Exception):
    pass


class ScreenLogicWarning(Exception):
    pass


class MESSAGE:
    COM_TIMEOUT = 5
    HEADER_FORMAT = "<HHI"
    HEADER_LENGTH = struct.calcsize(HEADER_FORMAT)


# Some of the message codes
# ANSWERS are QUERY + 1
class CODE:
    MSG_CODE_1 = 0
    UNKNOWN_ANSWER = 13
    CHALLENGE_QUERY = 14
    PING_QUERY = 16
    LOCALLOGIN_QUERY = 27
    VERSION_QUERY = 8120
    WEATHER_FORECAST_CHANGED = 9806
    WEATHER_FORECAST_QUERY = 9807
    STATUS_CHANGED = 12500
    COLOR_UPDATE = 12504
    CHEMISTRY_CHANGED = 12505
    ADD_CLIENT_QUERY = 12522
    REMOVE_CLIENT_QUERY = 12524
    POOLSTATUS_QUERY = 12526
    SETHEATTEMP_QUERY = 12528
    BUTTONPRESS_QUERY = 12530
    CTRLCONFIG_QUERY = 12532
    SETHEATMODE_QUERY = 12538
    LIGHTCOMMAND_QUERY = 12556
    SETCHEMDATA_QUERY = 12594
    EQUIPMENT_QUERY = 12566
    SCGCONFIG_QUERY = 12572
    SETSCG_QUERY = 12576
    PUMPSTATUS_QUERY = 12584
    SETCOOLTEMP_QUERY = 12590
    CHEMISTRY_QUERY = 12592
    GATEWAYDATA_QUERY = 18003


class DATA:
    KEY_ALERTS = "alerts"
    KEY_BODIES = "bodies"
    KEY_CHEMISTRY = "chemistry"
    KEY_CIRCUITS = "circuits"
    KEY_COLORS = "colors"
    KEY_CONFIG = "config"
    KEY_NOTIFICATIONS = "notifications"
    KEY_PUMPS = "pumps"
    KEY_SCG = "scg"
    KEY_SENSORS = "sensors"


class BODY_TYPE:
    POOL = 0
    SPA = 1

    NAME_FOR_NUM = {POOL: "Pool", SPA: "Spa"}

    NUM_FOR_NAME = {name: num for num, name in NAME_FOR_NUM.items()}


class ON_OFF:
    OFF = 0
    ON = 1

    NAME_FOR_NUM = {OFF: "Off", ON: "On"}

    NUM_FOR_NAME = {name: num for num, name in NAME_FOR_NUM.items()}

    @classmethod
    def from_bool(cls, expresson: bool):
        return cls.ON if expresson else cls.OFF


class HEAT_MODE:
    OFF = 0
    SOLAR = 1
    SOLAR_PREFERRED = 2
    HEATER = 3
    DONT_CHANGE = 4

    NAME_FOR_NUM = {
        OFF: "Off",
        SOLAR: "Solar",
        SOLAR_PREFERRED: "Solar Preferred",
        HEATER: "Heater",
        DONT_CHANGE: "Don't Change",
    }

    NUM_FOR_NAME = {name: num for num, name in NAME_FOR_NUM.items()}


class COLOR_MODE:
    OFF = 0
    ON = 1
    SET = 2
    SYNC = 3
    SWIM = 4
    PARTY = 5
    ROMANCE = 6
    CARIBBEAN = 7
    AMERICAN = 8
    SUNSET = 9
    ROYAL = 10
    SAVE = 11
    RECALL = 12
    BLUE = 13
    GREEN = 14
    RED = 15
    WHITE = 16
    MAGENTA = 17
    THUMPER = 18
    NEXT = 19
    RESET = 20
    HOLD = 21

    NAME_FOR_NUM = {
        OFF: "All Off",
        ON: "All On",
        SET: "Color Set",
        SYNC: "Color Sync",
        SWIM: "Color Swim",
        PARTY: "Party",
        ROMANCE: "Romance",
        CARIBBEAN: "Caribbean",
        AMERICAN: "American",
        SUNSET: "Sunset",
        ROYAL: "Royal",
        SAVE: "Save",
        RECALL: "Recall",
        BLUE: "Blue",
        GREEN: "Green",
        RED: "Red",
        WHITE: "White",
        MAGENTA: "Magenta",
        THUMPER: "Thumper",
        NEXT: "Next Mode",
        RESET: "Reset",
        HOLD: "Hold",
    }

    NUM_FOR_NAME = {name: num for num, name in NAME_FOR_NUM.items()}


class RANGE:
    MIN = 0
    MAX = 1

    NAME_FOR_NUM = {MIN: "Minimum", MAX: "Maximum"}


class DEVICE_TYPE:
    ALARM = "alarm"
    DURATION = "duration"
    ENERGY = "energy"
    ENUM = "enum"
    POWER = "power"
    TEMPERATURE = "temperature"
    VOLUME = "volume"


class STATE_TYPE:
    MEASUREMENT = "measurement"
    TOTAL = "total"
    TOTAL_INCREASING = "total_increasing"


class UNIT:
    # Chemistry
    PARTS_PER_MILLION = "ppm"
    PERCENT = "%"
    PH = "pH"
    SATURATION_INDEX = "lsi"

    # Electrical
    MILLIVOLT = "mV"
    WATT = "W"

    # Flow
    GALLONS_PER_MINUTE = "gpm"

    # Speed
    REVOLUTIONS_PER_MINUTE = "rpm"

    # Temperature
    CELSIUS = "\xb0C"
    FAHRENHEIT = "\xb0F"

    # Time
    HOUR = "hr"
    SECOND = "sec"

    # Volume
    MILLILITER = "mL"


class EQUIPMENT:
    CONTROLLER_HARDWARE = {
        0: {0: "IntelliTouch i5+3S"},
        1: {0: "IntelliTouch i7+3"},
        2: {0: "IntelliTouch i9+3"},
        3: {0: "IntelliTouch i5+3S"},
        4: {0: "IntelliTouch i9+3S"},
        5: {0: "IntelliTouch i10+3D", 1: "IntelliTouch i10X"},
        11: {0: "Suntouch/Intellicom"},
        13: {
            0: "EasyTouch2 8",
            1: "EasyTouch2 8P",
            2: "EasyTouch2 4",
            3: "EasyTouch2 4P",
            5: "EasyTouch2 PL4",
            6: "EasyTouch2 PSL4",
        },
        14: {
            0: "EasyTouch1 8",
            1: "EasyTouch1 8P",
            2: "EasyTouch1 4",
            3: "EasyTouch1 4P",
        },
    }

    PUMP_TYPE = {
        0: "None",
        1: "Intelliflo VF",
        2: "Intelliflo VS",
        3: "Intelliflo VSF",
    }

    FLAG_SOLAR = 0x1
    FLAG_SOLAR_AS_HEAT_PUMP = 0x2
    FLAG_CHLORINATOR = 0x4
    FLAG_SPA_SIDE_REMOTE = 0x20
    FLAG_COOLING = 0x800
    FLAG_INTELLICHEM = 0x8000


class CIRCUIT_FUNCTION:
    # Known circuit functions.
    GENERIC = 0
    SPA = 1
    POOL = 2
    MASTER_CLEANER = 5
    LIGHT = 7
    DIMMER = 8
    SAM_LIGHT = 9
    SAL_LIGHT = 10
    PHOTONGEN = 11
    COLOR_WHEEL = 12
    VALVE = 13
    SPILLWAY = 14
    FLOOR_CLEANER = 15
    INTELLIBRITE = 16
    MAGICSTREAM = 17

    GROUP_CORE = {
        SPA,
        POOL,
    }

    GROUP_LIGHTS_BASIC = {
        LIGHT,
        DIMMER,
    }

    GROUP_LIGHTS_COLOR = {
        SAM_LIGHT,
        SAL_LIGHT,
        PHOTONGEN,
        COLOR_WHEEL,
        INTELLIBRITE,
        MAGICSTREAM,
    }

    GROUP_LIGHTS_ALL = {
        *GROUP_LIGHTS_BASIC,
        *GROUP_LIGHTS_COLOR,
    }


class INTERFACE_GROUP:
    # Known interface groups
    POOL = 0
    SPA = 1
    FEATURES = 2
    LIGHTS_COLOR = 3  # ?
    LIGHTS = 4
    DONT_SHOW = 5


class CHEMISTRY:
    FLAG_ALARM_FLOW = 0x01
    FLAG_ALARM_PH = 0x06
    FLAG_ALARM_ORP = 0x18
    FLAG_ALARM_PH_SUPPLY = 0x20
    FLAG_ALARM_ORP_SUPPLY = 0x40
    FLAG_ALARM_PROBE_FAULT = 0x80

    # Unconfirmed
    FLAG_STATUS_NORMAL = 0x01
    FLAG_STATUS_IDEAL = 0x02
    FLAG_STATUS_CORROSIVE = 0x04
    FLAG_STATUS_SCALING = 0x08

    MASK_STATUS_ORP_DOSE_TYPE = 0x03
    MASK_STATUS_PH_DOSE_TYPE = 0x0C
    MASK_STATUS_PH_DOSING = 0x30
    MASK_STATUS_ORP_DOSING = 0xC0

    FLAG_WARNING_PH_LOCKOUT = 0x01
    FLAG_WARNING_PH_LIMIT = 0x02
    FLAG_WARNING_ORP_LIMIT = 0x04
    FLAG_WARNING_INVALID_SETUP = 0x08
    FLAG_WARNING_CHLORINATOR_COMM_ERROR = 0x10

    FLAG_FLAGS_FLOW_DELAY = 0x02
    FLAG_FLAGS_INTELLICHLOR = 0x04
    FLAG_FLAGS_MANUAL_DOSING = 0x08
    FLAG_FLAGS_USE_CHLORINATOR = 0x10
    FLAG_FLAGS_ADVANCED_DISPLAY = 0x20
    FLAG_FLAGS_PH_SUPPLY_TYPE = 0x40
    FLAG_FLAGS_COMMS_LOST = 0x80  # ?

    # Valid ranges listed in IntelliChem documentation
    RANGE_PH_SETPOINT = {RANGE.MIN: 7.2, RANGE.MAX: 7.6}
    RANGE_ORP_SETPOINT = {RANGE.MIN: 400, RANGE.MAX: 800}


class CHEM_DOSING_STATE:
    DOSING = 0
    MIXING = 1
    MONITORING = 2

    NAME_FOR_NUM = {DOSING: "Dosing", MIXING: "Mixing", MONITORING: "Monitoring"}
    NUM_FOR_NAME = {name: num for num, name in NAME_FOR_NUM.items()}


class CHEM_DOSE_TYPE:
    class ORP:
        NONE = 0
        CHLORINE = 1
        SCG = 2  # ?

        NAME_FOR_NUM = {
            NONE: "None",
            CHLORINE: "Chlorine",
            SCG: "Salt Chlorine Generator",
        }
        NUM_FOR_NAME = {name: num for num, name in NAME_FOR_NUM.items()}

    class PH:
        NONE = 0
        ACID = 1
        CO2 = 2

        NAME_FOR_NUM = {
            NONE: "None",
            ACID: "Acid",
            CO2: "CO2",
        }
        NUM_FOR_NAME = {name: num for num, name in NAME_FOR_NUM.items()}


class SCG:
    LIMIT_FOR_BODY = {BODY_TYPE.POOL: 100, BODY_TYPE.SPA: 100}
    MAX_SC_RUNTIME = 72


GENERIC_CIRCUIT_NAMES = [
    *[f"Aux {num}" for num in range(1, 25)],  # Last number is 24
    "AuxEx",
    *[f"Feature {num}" for num in range(1, 9)],  # Last number is 8
]

DEFAULT_CIRCUIT_NAMES = ["Spa", "Pool", *GENERIC_CIRCUIT_NAMES]


# COLOR_MODES_* may not be complete
COLOR_MODES_GENERIC = {
    num: COLOR_MODE.NAME_FOR_NUM[num]
    for num in [
        COLOR_MODE.OFF,
        COLOR_MODE.ON,
    ]
}

COLOR_MODES_COLORS = {
    num: COLOR_MODE.NAME_FOR_NUM[num]
    for num in [
        COLOR_MODE.BLUE,
        COLOR_MODE.GREEN,
        COLOR_MODE.RED,
        COLOR_MODE.WHITE,
        COLOR_MODE.MAGENTA,
    ]
}

COLOR_MODES_SAM = {
    **COLOR_MODES_GENERIC,
    **{
        num: COLOR_MODE.NAME_FOR_NUM[num]
        for num in [
            COLOR_MODE.SET,
            COLOR_MODE.SYNC,
            COLOR_MODE.SWIM,
        ]
    },
    **COLOR_MODES_COLORS,
}

COLOR_MODES_INTELLIBRITE = {
    **COLOR_MODES_GENERIC,
    **COLOR_MODES_SAM,
    **{
        num: COLOR_MODE.NAME_FOR_NUM[num]
        for num in [
            COLOR_MODE.PARTY,
            COLOR_MODE.ROMANCE,
            COLOR_MODE.CARIBBEAN,
            COLOR_MODE.AMERICAN,
            COLOR_MODE.SUNSET,
            COLOR_MODE.ROYAL,
            COLOR_MODE.SAVE,
            COLOR_MODE.RECALL,
        ]
    },
    **COLOR_MODES_COLORS,
}

COLOR_MODES_MAGICSTREAM = {
    **COLOR_MODES_GENERIC,
    **{
        num: COLOR_MODE.NAME_FOR_NUM[num]
        for num in [
            COLOR_MODE.THUMPER,
            COLOR_MODE.NEXT,
            COLOR_MODE.RESET,
            COLOR_MODE.HOLD,
        ]
    },
}



import struct

from ..const import CODE
from .protocol import ScreenLogicProtocol
from .request import async_make_request


async def async_request_pool_button_press(
    protocol: ScreenLogicProtocol, circuit_id: int, circuit_state: int
) -> bool:
    return (
        await async_make_request(
            protocol,
            CODE.BUTTONPRESS_QUERY,
            struct.pack("<III", 0, circuit_id, circuit_state),
        )
        == b""
    )



from ..const import CODE
from .protocol import ScreenLogicProtocol
from .request import async_make_request
from .utility import decodeMessageString


async def async_request_gateway_version(protocol: ScreenLogicProtocol):
    if result := await async_make_request(protocol, CODE.VERSION_QUERY):
        return decodeMessageString(result)



import asyncio
import sys
from screenlogicpy.cli import cli


def main():
    args = sys.argv[1:]
    # Save for debugger
    # args = ["-i", "xx.xx.xx.xx", "get", "json"]
    if sys.platform == "win32":
        asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
    return asyncio.run(cli(args))


if __name__ == "__main__":
    sys.exit(main())



from ..const import DATA
from .utility import getSome, getString


def decode_color_update(buff: bytes, data: dict):
    config = data.setdefault(DATA.KEY_CONFIG, {})

    color_state = config.setdefault("color_state", {})

    mode, offset = getSome("I", buff, 0)  # 0
    color_state["mode"] = mode

    progress, offset = getSome("I", buff, offset)  # 4
    color_state["progress"] = progress

    limit, offset = getSome("I", buff, offset)  # 8
    color_state["limit"] = limit

    text, offset = getString(buff, offset)  # 12
    color_state["text"] = text



import asyncio
import logging
import struct
from typing import Callable

from ..const import CODE, MESSAGE, ScreenLogicError
from .protocol import ScreenLogicProtocol
from .utility import asyncio_timeout, decodeMessageString, encodeMessageString

_LOGGER = logging.getLogger(__name__)


def create_login_message():
    # these constants are only for this message.
    schema = 348
    connectionType = 0
    clientVersion = encodeMessageString("Android")
    pid = 2
    password = "0000000000000000"  # passwd must be <= 16 chars. empty is not OK.
    passwd = encodeMessageString(password)
    fmt = "<II" + str(len(clientVersion)) + "s" + str(len(passwd)) + "sxI"
    return struct.pack(fmt, schema, connectionType, clientVersion, passwd, pid)


def create_local_login_message():
    schema = 348
    connectionType = 0
    clientVersion = encodeMessageString("Local Config")
    passwdPayload = b"\x10\x00\x00\x00\x48\x9e\x60\x3a\xc3\x1d\xb9\xb1\x0c\xc1\x4a\x37\x50\x97\xa8\x22"
    mac = encodeMessageString("00-00-00-00-00-00")
    pid = 2
    fmt = f"<II{len(clientVersion)}s{len(passwdPayload)}sI{len(mac)}sI"
    return struct.pack(
        fmt, schema, connectionType, clientVersion, passwdPayload, pid, mac, 0
    )


async def async_get_mac_address(gateway_ip, gateway_port):
    """Connect to a screenlogic gateway and return the mac address only."""
    transport, protocol = await async_create_connection(gateway_ip, gateway_port)
    mac = await async_gateway_connect(transport, protocol)
    if transport and not transport.is_closing():
        transport.close()
    return mac


async def async_create_connection(
    gateway_ip, gateway_port, connection_lost_callback: Callable = None
):
    try:
        loop = asyncio.get_running_loop()

        # on_con_lost = loop.create_future()
        _LOGGER.debug("Creating connection")
        async with asyncio_timeout(MESSAGE.COM_TIMEOUT):
            return await loop.create_connection(
                lambda: ScreenLogicProtocol(loop, connection_lost_callback),
                gateway_ip,
                gateway_port,
            )
    except (OSError, asyncio.TimeoutError) as ex:
        raise ScreenLogicError(
            f"Failed to connect to host at {gateway_ip}:{gateway_port}"
        ) from ex


async def async_gateway_connect(
    transport: asyncio.Transport, protocol: ScreenLogicProtocol
) -> str:
    connectString = b"CONNECTSERVERHOST\r\n\r\n"  # as bytes, not string
    try:
        # Connect ping
        _LOGGER.debug("Pinging protocol adapter")
        transport.write(connectString)
    except Exception as ex:
        raise ScreenLogicError("Error sending connect ping") from ex

    await asyncio.sleep(0.25)
    if not protocol.is_connected:
        raise ScreenLogicError("Host unexpectedly disconnected.")

    _LOGGER.debug("Sending challenge")
    request = protocol.await_send_message(CODE.CHALLENGE_QUERY)

    try:

        async with asyncio_timeout(MESSAGE.COM_TIMEOUT):
            await request
    except asyncio.TimeoutError:
        raise ScreenLogicError("Host failed to respond to challenge")

    if not request.cancelled():
        # mac address
        return decodeMessageString(request.result())


async def async_gateway_login(protocol: ScreenLogicProtocol) -> bool:
    _LOGGER.debug("Logging in")
    request = protocol.await_send_message(CODE.LOCALLOGIN_QUERY, create_login_message())
    try:
        async with asyncio_timeout(MESSAGE.COM_TIMEOUT):
            await request
    except asyncio.TimeoutError:
        raise ScreenLogicError("Failed to logon to gateway")

    return not request.cancelled()


async def async_connect_to_gateway(
    gateway_ip, gateway_port, connection_lost_callback: Callable = None
):
    transport, protocol = await async_create_connection(
        gateway_ip, gateway_port, connection_lost_callback
    )
    mac_address = await async_gateway_connect(transport, protocol)
    if await async_gateway_login(protocol):
        return transport, protocol, mac_address



"""Client manager for a connection to a ScreenLogic protocol adapter."""
import asyncio
import logging
from typing import Callable

from .const import CODE, COM_KEEPALIVE, ScreenLogicWarning
from .requests.chemistry import decode_chemistry
from .requests.client import async_request_add_client, async_request_remove_client
from .requests.color import decode_color_update
from .requests.status import decode_pool_status
from .requests.ping import async_request_ping
from .requests.protocol import ScreenLogicProtocol

_LOGGER = logging.getLogger(__name__)


class ClientManager:
    """Class to manage callback subscriptions to specific ScreenLogic messages."""

    def __init__(self) -> None:
        self._listeners = {}
        self._is_client = False
        self._client_sub_unsub_lock = asyncio.Lock()
        self._protocol = None
        self._data = None

    @property
    def is_client(self):
        """Return if connected to ScreenLogic as a client."""
        return self._is_client and self._protocol and self._protocol.is_connected

    @property
    def client_needed(self):
        """Return if desired to be a client."""
        return self._listeners and not self._is_client

    def _attached(self):
        return self._protocol and self._protocol.is_connected

    async def attach(self, protocol: ScreenLogicProtocol, data: dict):
        """
        Update protocol and data reference.

        Updates this ClientManager's reference to a ScreenLogicProtocol instance
        and a current data dict. Will attempt to re-register any existing callbacks
        to the new protocol instance.
        """
        self._protocol = protocol
        self._data = data
        self._is_client = False
        if self.client_needed:
            self._protocol.remove_all_async_message_callbacks()
            for code in self._listeners:
                self._protocol.register_async_message_callback(
                    code, self._async_common_callback, code, self._data
                )
            await self.async_subscribe_gateway()

    def _notify_listeners(self, code: int) -> None:
        """Notify all listeners."""
        for callback in self._listeners.get(code):
            callback()

    def _callback_factory(self, code) -> Callable:
        """Return decoding method for known message codes."""
        if code == CODE.STATUS_CHANGED:
            return decode_pool_status
        elif code == CODE.CHEMISTRY_CHANGED:
            return decode_chemistry
        elif code == CODE.COLOR_UPDATE:
            return decode_color_update
        else:
            return None

    async def _async_common_callback(self, message, code, data):
        """Decode known incoming messages."""
        if decoder := self._callback_factory(code):
            decoder(message, data)

        self._notify_listeners(code)

    async def async_subscribe(
        self, callback: Callable[..., any], code: int
    ) -> Callable:
        """
        Register listener callback.

        Registers a callback method to call when a message with the specified
        message code is received. Messages with known codes will be processed
        and applied to gateway data before callback method is called.
        """
        if not self._attached():
            return None

        _LOGGER.debug(f"Adding listener {callback}")
        code_listeners: set = self._listeners.setdefault(code, set())

        code_listeners.add(callback)

        if self._attached():
            self._protocol.register_async_message_callback(
                code, self._async_common_callback, code, self._data
            )

        if self.client_needed:
            _LOGGER.debug("Client needed.")
            if not await self.async_subscribe_gateway():
                return None

        def remove_listener():
            """Remove listener callback."""
            if callback in code_listeners:
                _LOGGER.debug(f"Removing listener {callback}")
                code_listeners.remove(callback)
                if not code_listeners:
                    _LOGGER.debug(f"No more listeners for code {code}. Removing.")
                    if code in self._listeners:
                        self._listeners.pop(code)
                        if self._attached():
                            self._protocol.remove_async_message_callback(code)
                            if not self._listeners:
                                _LOGGER.debug(
                                    "No more listeners for any code. Unsubscribing gateway."
                                )
                                asyncio.create_task(self.async_unsubscribe_gateway())

        return remove_listener

    async def _async_ping(self):
        """Check connection before requesting a ping."""
        if not self._attached():
            raise ScreenLogicWarning(
                "Not attached to protocol adapter. request_ping failed."
            )
        _LOGGER.debug("Requesting ping")
        if await async_request_ping(self._protocol):
            _LOGGER.debug("Ping successful.")

    async def _async_add_client(self):
        """Check connection before sending add client request."""
        if not self._attached():
            raise ScreenLogicWarning(
                "Not attached to protocol adapter. add_client failed."
            )
        _LOGGER.debug("Requesting add client")
        return await async_request_add_client(self._protocol)

    async def _async_remove_client(self):
        """Check connection before sending remove client request."""
        if not self._attached():
            raise ScreenLogicWarning(
                "Not attached to protocol adapter. remove_client failed."
            )
        _LOGGER.debug("Requesting remove client")
        return await async_request_remove_client(self._protocol)

    async def async_subscribe_gateway(self) -> bool:
        """
        Subscribe as ScreenLogic client.

        Adds this gateway as a client on the ScreenLogic protocol adapter. This
        tells ScreenLogic that we are interested in receiving push update messages.
        """
        if self._attached():
            async with self._client_sub_unsub_lock:
                if not self.is_client:
                    _LOGGER.debug("Subscribing gateway.")
                    if await self._async_add_client():
                        self._is_client = True
                        self._protocol.enable_keepalive(self._async_ping, COM_KEEPALIVE)
                        _LOGGER.debug("Gateway subscribed")
                        return True
                    return False
                return True

    async def async_unsubscribe_gateway(self) -> bool:
        """
        Unsubscribe as ScreenLogic client.

        Removes this gateway as a client on the ScreenLogic protocol adapter.
        ScreenLogic will no longer push update messages.
        """
        if self._attached():
            async with self._client_sub_unsub_lock:
                if self._is_client:
                    self._is_client = False
                    self._protocol.disable_keepalive()
                    self._protocol.remove_all_async_message_callbacks()
                    _LOGGER.debug("Gateway unsubscribing")
                    return await self._async_remove_client()
                return True



import pytest

from screenlogicpy.requests.login import async_connect_to_gateway, async_get_mac_address
from tests.fake_gateway import (
    FAKE_GATEWAY_ADDRESS,
    FAKE_GATEWAY_MAC,
    FAKE_GATEWAY_PORT,
)


@pytest.mark.asyncio
async def test_async_gateway_login(MockProtocolAdapter):
    async with MockProtocolAdapter:
        _, _, mac_address = await async_connect_to_gateway(
            FAKE_GATEWAY_ADDRESS, FAKE_GATEWAY_PORT
        )

    assert mac_address == FAKE_GATEWAY_MAC


@pytest.mark.asyncio
async def test_async_get_mac_address(MockProtocolAdapter):
    async with MockProtocolAdapter:
        mac_address = await async_get_mac_address(
            FAKE_GATEWAY_ADDRESS, FAKE_GATEWAY_PORT
        )

    assert mac_address == FAKE_GATEWAY_MAC



"""Defines an asyncio.Protocol for communicating with Pentair ScreenLogic systems."""
import asyncio
import itertools
import logging
import struct
import time
from typing import Awaitable, Callable, Tuple, List

from ..const import CODE, MESSAGE, ScreenLogicError
from .utility import makeMessage, takeMessage

_LOGGER = logging.getLogger(__name__)


class ScreenLogicProtocol(asyncio.Protocol):
    """asyncio.Protocol for handling connection to a ScreenLogic protocol adapter."""

    def __init__(self, loop, connection_lost_callback: Callable = None) -> None:
        self._loop: asyncio.BaseEventLoop = loop
        self._connection_lost_callback = connection_lost_callback
        self._futures = self.FutureManager(self._loop)
        self._callbacks = {}
        self._connected = False
        self._last_request: float = None
        self._last_response: float = None
        self._buff = bytearray()

        self._keepalive_awaitable: Callable[[any, any], Awaitable[any]] = None
        self._keepalive_interval: int = None
        self._stop_keepalive: Callable = None

        # Adapter-initiated message IDs seem to start at 32767,
        # so we'll use only the lower half of the message ID data size.
        self.__msgID = itertools.cycle(range(32767))

    @property
    def is_connected(self):
        """Return if protocol is currently connected."""
        return self._connected

    @property
    def last_request(self):
        """Monotonic time for the last message sent."""
        return self._last_request

    @property
    def last_response(self):
        """Monotonic time for last message received."""
        return self._last_response

    def connection_made(self, transport: asyncio.Transport) -> None:
        """Called when connection is made."""
        self._connected = True
        _LOGGER.debug("Connected to server")
        self.transport = transport

    def send_message(self, messageID, messageCode, messageData=b"") -> None:
        """Send a message via the transport."""
        _LOGGER.debug("Sending: %i, %i, %s", messageID, messageCode, messageData)
        self.transport.write(makeMessage(messageID, messageCode, messageData))

        self._last_request = time.monotonic()

        if self._keepalive_awaitable:
            self._set_keepalive()

    def await_send_message(self, messageCode, messageData=b"") -> asyncio.Future:
        """
        Send a message and return an awaitable.

        Sends the message and returns an awaitable asyncio.Future object that will
        contain the result of the ScreenLogic protocol adapter's response.
        """
        messageID = next(self.__msgID)
        fut = self._futures.create(messageID)
        self.send_message(messageID, messageCode, messageData)
        return fut

    def data_received(self, data: bytes) -> None:
        """Called with data is received."""

        def complete_messages(data: bytes) -> List[Tuple[int, int, bytes]]:
            """Return only complete ScreenLogic messages."""

            # Some pool configurations can require SL messages larger than can
            # come through in a single call to data_received(), so lets wait until
            # we have at least enough data to make a complete message before we
            # process and pass it on. Conversely, multiple SL messages may come in
            # a single call to data_received() so we collect all complete messages
            # before sending on.

            self._buff.extend(data)
            complete = []
            while len(self._buff) >= MESSAGE.HEADER_LENGTH:
                dataLen = struct.unpack_from("<I", self._buff, 4)[0]
                totalLen = MESSAGE.HEADER_LENGTH + dataLen
                if len(self._buff) >= totalLen:
                    out = bytearray()
                    for _ in range(totalLen):
                        out.append(self._buff.pop(0))
                    complete.append(takeMessage(bytes(out)))
                else:
                    break
            if len(self._buff) > 0:
                _LOGGER.debug(
                    f"Returning {len(complete)} messages with {len(self._buff)} bytes in the buffer"
                )
                _LOGGER.debug(f"Buffer: {self._buff}")
            return complete

        for messageID, messageCode, message in complete_messages(data):

            if messageCode == CODE.UNKNOWN_ANSWER:
                raise ScreenLogicError(f"Request explicitly rejected: {messageID}")

            if self._futures.mark_done(messageID, message):
                _LOGGER.debug("Received: %i, %i, %s", messageID, messageCode, message)
            else:
                _LOGGER.debug(
                    "Received async message: %i, %i, %s",
                    messageID,
                    messageCode,
                    message,
                )
                # Unsolicited message received. See if there's a callback registered
                # for the message code and create a task for it.
                if messageCode in self._callbacks:
                    handler, args = self._callbacks[messageCode]
                    _LOGGER.debug(f"Calling {handler}")
                    self._loop.create_task(handler(message, *args))

    def connection_lost(self, exc) -> None:
        """Called when connection is closed/lost."""
        self._connected = False
        _LOGGER.debug("Connection closed")
        if self._connection_lost_callback is not None:
            self._connection_lost_callback()

    def register_async_message_callback(
        self,
        messageCode,
        handler: Callable[[bytes, any], Awaitable[any]],
        *args,
    ) -> None:
        """
        Register callback for async ScreenLogic message.

        Registers an async callback function to call for the specified message code.
        Callback will be scheduled to run with loop.create_task()
        """
        _LOGGER.debug(
            f"Registering async handler {handler} for message code {messageCode}"
        )
        self._callbacks[messageCode] = (handler, args)

    def remove_async_message_callback(self, messageCode) -> bool:
        """Remove callback for message code."""
        return True if self._callbacks.pop(messageCode, None) else False

    def remove_all_async_message_callbacks(self) -> None:
        """Remove all message callbacks."""
        self._callbacks.clear()

    def requests_pending(self) -> bool:
        """Return if requests are pending."""
        return not self._futures.all_done()

    def _call_keepalive(self) -> None:
        """Schedule keepalive callback."""
        _LOGGER.debug("Creating keepalive task")
        task = self._loop.create_task(self._keepalive_awaitable())
        _LOGGER.debug(f"keepalive task {task} created")

    def _set_keepalive(self) -> None:
        """Call keepalive scheduler after set delay."""
        if self._stop_keepalive:
            _LOGGER.debug("Killing current keepalive")
            self._stop_keepalive()
            self._stop_keepalive = None

        handle: asyncio.TimerHandle

        _LOGGER.debug("Setting keepalive call_later")
        handle = self._loop.call_later(self._keepalive_interval, self._call_keepalive)

        def _stop():
            """Stop current keepalive timer."""
            if handle:
                _LOGGER.debug(f"Canceling call_later {handle}")
                handle.cancel()

        _LOGGER.debug("Assigning stop_keepalilve callback")
        self._stop_keepalive = _stop

    def enable_keepalive(
        self,
        keepalive_awaitable: Callable[[any, any], Awaitable[any]],
        keepalive_interval: int,
    ) -> None:
        """Enable connection keepalive."""
        _LOGGER.debug("Enabling keepalive")
        self._keepalive_awaitable = keepalive_awaitable
        self._keepalive_interval = keepalive_interval
        self._set_keepalive()

    def disable_keepalive(self) -> None:
        """Disable connection keepalive"""
        _LOGGER.debug("Disabling keepalive")
        self._keepalive_awaitable = None
        if self._stop_keepalive:
            self._stop_keepalive()
            self._stop_keepalive = None

    class FutureManager:
        """Class to manage responses for pending ScreenLogic requests."""

        def __init__(self, loop: asyncio.AbstractEventLoop) -> None:
            self._collection = {}
            self.loop = loop

        def create(self, msgID) -> asyncio.Future:
            """Create future for response."""
            self._collection[msgID] = self.loop.create_future()
            return self._collection[msgID]

        def try_get(self, msgID) -> asyncio.Future:
            """Get response future for message ID."""
            fut: asyncio.Future
            if (fut := self._collection.pop(msgID, None)) is not None:
                if not fut.cancelled():
                    return fut
            return None

        def mark_done(self, msgID, result=True) -> bool:
            """Mark future done and add response."""
            if (fut := self.try_get(msgID)) is not None:
                try:
                    fut.set_result(result)
                except asyncio.exceptions.InvalidStateError as ise:
                    raise ScreenLogicError(
                        f"Attempted to set result on future {msgID} when result exists: {fut.result()}"
                    ) from ise
                return True
            return False

        def all_done(self) -> bool:
            """Return if outstanding still futures exist."""
            fut: asyncio.Future
            for fut in self._collection.values():
                if not fut.done():
                    return False
            return True



import struct

from ..const import CODE, DATA, STATE_TYPE, UNIT
from .protocol import ScreenLogicProtocol
from .request import async_make_request
from .utility import getSome


async def async_request_scg_config(protocol: ScreenLogicProtocol, data: dict) -> bytes:
    if result := await async_make_request(
        protocol, CODE.SCGCONFIG_QUERY, struct.pack("<I", 0)
    ):
        decode_scg_config(result, data)
        return result


def decode_scg_config(buff: bytes, data: dict) -> None:
    scg = data.setdefault(DATA.KEY_SCG, {})

    present, offset = getSome("I", buff, 0)  # 0
    scg["scg_present"] = present

    status, offset = getSome("I", buff, offset)  # 4
    scg["scg_status"] = {
        "name": "SCG Status",
        "value": status,
    }

    level1, offset = getSome("I", buff, offset)  # 8
    scg["scg_level1"] = {
        "name": "Pool SCG Level",
        "value": level1,
        "unit": UNIT.PERCENT,
    }

    level2, offset = getSome("I", buff, offset)  # 12
    scg["scg_level2"] = {
        "name": "Spa SCG Level",
        "value": level2,
        "unit": UNIT.PERCENT,
    }

    salt, offset = getSome("I", buff, offset)  # 16
    scg["scg_salt_ppm"] = {
        "name": "SCG Salt",
        "value": (salt * 50),
        "unit": UNIT.PARTS_PER_MILLION,
        "state_type": STATE_TYPE.MEASUREMENT,
    }

    flags, offset = getSome("I", buff, offset)  # 20
    scg["scg_flags"] = flags

    superChlorTimer, offset = getSome("I", buff, offset)  # 24
    scg["scg_super_chlor_timer"] = {
        "name": "SCG Super Chlorination Timer",
        "value": superChlorTimer,
        "unit": UNIT.HOUR,
    }


async def async_request_set_scg_config(
    protocol: ScreenLogicProtocol,
    pool_output: int,
    spa_output: int,
    super_chlor: int = 0,
    super_time: int = 0,
) -> bool:
    return (
        await async_make_request(
            protocol,
            CODE.SETSCG_QUERY,
            struct.pack("<IIIII", 0, pool_output, spa_output, super_chlor, super_time),
        )
        == b""
    )



# import json
import struct

from ..const import (
    CODE,
    BODY_TYPE,
    DATA,
    DEVICE_TYPE,
    ON_OFF,
    STATE_TYPE,
    UNIT,
)
from .protocol import ScreenLogicProtocol
from .request import async_make_request
from .utility import getSome, getTemperatureUnit


async def async_request_pool_status(protocol: ScreenLogicProtocol, data: dict) -> bytes:
    if result := await async_make_request(
        protocol, CODE.POOLSTATUS_QUERY, struct.pack("<I", 0)
    ):
        decode_pool_status(result, data)
        return result


def decode_pool_status(buff: bytes, data: dict) -> None:
    config = data.setdefault(DATA.KEY_CONFIG, {})

    ok, offset = getSome("I", buff, 0)  # byte offset 0
    config["ok"] = ok

    freezeMode, offset = getSome("B", buff, offset)  # byte offset 4
    config["freeze_mode"] = {
        "name": "Freeze Mode",
        "value": ON_OFF.from_bool(freezeMode & 0x08),
    }

    remotes, offset = getSome("B", buff, offset)  # 5
    config["remotes"] = {"name": "Remotes", "value": remotes}

    poolDelay, offset = getSome("B", buff, offset)  # 6
    config["pool_delay"] = {"name": "Pool Delay", "value": poolDelay}

    spaDelay, offset = getSome("B", buff, offset)  # 7
    config["spa_delay"] = {"name": "Spa Delay", "value": spaDelay}

    cleanerDelay, offset = getSome("B", buff, offset)  # 8
    config["cleaner_delay"] = {"name": "Cleaner Delay", "value": cleanerDelay}

    config[f"unknown_at_offset_{offset:02}"], offset = getSome("B", buff, offset)  # 9
    config[f"unknown_at_offset_{offset:02}"], offset = getSome("B", buff, offset)  # 10
    config[f"unknown_at_offset_{offset:02}"], offset = getSome("B", buff, offset)  # 11

    sensors = data.setdefault(DATA.KEY_SENSORS, {})

    temperature_unit = getTemperatureUnit(data)

    airTemp, offset = getSome("i", buff, offset)  # 12
    sensors["air_temperature"] = {
        "name": "Air Temperature",
        "value": airTemp,
        "unit": temperature_unit,
        "device_type": DEVICE_TYPE.TEMPERATURE,
        "state_type": STATE_TYPE.MEASUREMENT,
    }

    bodiesCount, offset = getSome("I", buff, offset)  # 16

    # Should this default to 2?
    bodiesCount = min(bodiesCount, 2)

    bodies: dict = data.setdefault(DATA.KEY_BODIES, {})

    for i in range(bodiesCount):
        currentBody: dict = bodies.setdefault(i, {})

        bodyType, offset = getSome("I", buff, offset)
        if bodyType not in range(2):
            bodyType = 0

        currentBody.setdefault("min_set_point", {})["unit"] = temperature_unit

        currentBody.setdefault("max_set_point", {})["unit"] = temperature_unit

        currentBody["body_type"] = {"name": "Type of body of water", "value": bodyType}

        lastTemp, offset = getSome("i", buff, offset)
        bodyName = "Last {} Temperature".format(BODY_TYPE.NAME_FOR_NUM[bodyType])
        currentBody["last_temperature"] = {
            "name": bodyName,
            "value": lastTemp,
            "unit": temperature_unit,
            "device_type": DEVICE_TYPE.TEMPERATURE,
            "state_type": STATE_TYPE.MEASUREMENT,
        }

        heatStatus, offset = getSome("i", buff, offset)
        heaterName = "{} Heat".format(BODY_TYPE.NAME_FOR_NUM[bodyType])
        currentBody["heat_status"] = {"name": heaterName, "value": heatStatus}

        heatSetPoint, offset = getSome("i", buff, offset)
        hspName = "{} Heat Set Point".format(BODY_TYPE.NAME_FOR_NUM[bodyType])
        currentBody["heat_set_point"] = {
            "name": hspName,
            "value": heatSetPoint,
            "unit": temperature_unit,
            "device_type": DEVICE_TYPE.TEMPERATURE,
        }

        coolSetPoint, offset = getSome("i", buff, offset)
        cspName = "{} Cool Set Point".format(BODY_TYPE.NAME_FOR_NUM[bodyType])
        currentBody["cool_set_point"] = {
            "name": cspName,
            "value": coolSetPoint,
            "unit": temperature_unit,
        }

        heatMode, offset = getSome("i", buff, offset)
        hmName = "{} Heat Mode".format(BODY_TYPE.NAME_FOR_NUM[bodyType])
        currentBody["heat_mode"] = {
            "name": hmName,
            "value": heatMode,
        }

    circuitCount, offset = getSome("I", buff, offset)

    circuits: dict = data.setdefault(DATA.KEY_CIRCUITS, {})

    for i in range(circuitCount):
        circuitID, offset = getSome("I", buff, offset)

        currentCircuit = circuits.setdefault(circuitID, {})

        if "id" not in currentCircuit:
            currentCircuit["id"] = circuitID

        circuitState, offset = getSome("I", buff, offset)
        currentCircuit["value"] = circuitState

        cColorSet, offset = getSome("B", buff, offset)
        currentCircuit["color_set"] = cColorSet

        cColorPos, offset = getSome("B", buff, offset)
        currentCircuit["color_position"] = cColorPos

        cColorStagger, offset = getSome("B", buff, offset)
        currentCircuit["color_stagger"] = cColorStagger

        circuitDelay, offset = getSome("B", buff, offset)
        currentCircuit["delay"] = circuitDelay

    pH, offset = getSome("i", buff, offset)
    sensors["ph"] = {
        "name": "pH",
        "value": (pH / 100),
        "unit": UNIT.PH,
        "state_type": STATE_TYPE.MEASUREMENT,
    }

    orp, offset = getSome("i", buff, offset)
    sensors["orp"] = {
        "name": "ORP",
        "value": orp,
        "unit": UNIT.MILLIVOLT,
        "state_type": STATE_TYPE.MEASUREMENT,
    }

    saturation, offset = getSome("i", buff, offset)
    sensors["saturation"] = {
        "name": "Saturation Index",
        "value": (saturation / 100),
        "unit": UNIT.SATURATION_INDEX,
        "state_type": STATE_TYPE.MEASUREMENT,
    }

    saltPPM, offset = getSome("i", buff, offset)
    sensors["salt_ppm"] = {
        "name": "Salt",
        "value": (saltPPM * 50),
        "unit": UNIT.PARTS_PER_MILLION,
        "state_type": STATE_TYPE.MEASUREMENT,
    }

    pHTank, offset = getSome("i", buff, offset)
    sensors["ph_supply_level"] = {
        "name": "pH Supply Level",
        "value": pHTank,
        "state_type": STATE_TYPE.MEASUREMENT,
    }

    orpTank, offset = getSome("i", buff, offset)
    sensors["orp_supply_level"] = {
        "name": "ORP Supply Level",
        "value": orpTank,
        "state_type": STATE_TYPE.MEASUREMENT,
    }

    alarm, offset = getSome("i", buff, offset)
    sensors["chem_alarm"] = {
        "name": "Chemistry Alarm",
        "value": alarm,
        "device_type": DEVICE_TYPE.ALARM,
    }



import asyncio
import pytest
import struct
from unittest.mock import patch

# from deepdiff import DeepDiff


from .const_data import (
    EXPECTED_COMPLETE_DATA,
    FAKE_CONNECT_INFO,
    FAKE_GATEWAY_ADDRESS,
    FAKE_GATEWAY_MAC,
    FAKE_GATEWAY_NAME,
    FAKE_GATEWAY_PORT,
    FAKE_GATEWAY_VERSION,
)
from screenlogicpy import ScreenLogicGateway


@pytest.mark.asyncio
async def test_gateway(MockConnectedGateway):
    gateway = MockConnectedGateway
    assert gateway.ip == FAKE_GATEWAY_ADDRESS
    assert gateway.port == FAKE_GATEWAY_PORT
    assert gateway.name == FAKE_GATEWAY_NAME
    assert gateway.version == FAKE_GATEWAY_VERSION
    assert gateway.mac == FAKE_GATEWAY_MAC
    data = gateway.get_data()
    await gateway.async_disconnect()
    assert data

    # diff = DeepDiff(data, EXPECTED_COMPLETE_DATA)
    # print(diff)
    assert data == EXPECTED_COMPLETE_DATA


@pytest.mark.asyncio
async def test_gateway_connect(MockProtocolAdapter):
    async with MockProtocolAdapter:
        gateway = ScreenLogicGateway(**FAKE_CONNECT_INFO)
        await gateway.async_connect()
        assert gateway.ip == FAKE_GATEWAY_ADDRESS
        assert gateway.port == FAKE_GATEWAY_PORT
        assert gateway.name == FAKE_GATEWAY_NAME
        assert gateway.version == FAKE_GATEWAY_VERSION
        assert gateway.mac == FAKE_GATEWAY_MAC
        assert gateway.is_connected
        await gateway.async_disconnect()


@pytest.mark.asyncio
async def test_gateway_late_connect(MockProtocolAdapter):
    async with MockProtocolAdapter:
        gateway = ScreenLogicGateway()
        await gateway.async_connect(**FAKE_CONNECT_INFO)
        assert gateway.ip == FAKE_GATEWAY_ADDRESS
        assert gateway.port == FAKE_GATEWAY_PORT
        assert gateway.name == FAKE_GATEWAY_NAME
        assert gateway.version == FAKE_GATEWAY_VERSION
        assert gateway.mac == FAKE_GATEWAY_MAC
        assert gateway.is_connected
        await gateway.async_disconnect()


@pytest.mark.asyncio
async def test_async_set_circuit(
    event_loop: asyncio.AbstractEventLoop, MockConnectedGateway
):
    circuit_id = 505
    circuit_state = 1
    result = event_loop.create_future()
    result.set_result(b"")
    with patch(
        "screenlogicpy.requests.button.ScreenLogicProtocol.await_send_message",
        return_value=result,
    ) as mockRequest:
        gateway = MockConnectedGateway
        assert await gateway.async_set_circuit(circuit_id, circuit_state)
        await gateway.async_disconnect()
        assert mockRequest.call_args.args[0] == 12530
        assert mockRequest.call_args.args[1] == struct.pack(
            "<III", 0, circuit_id, circuit_state
        )


@pytest.mark.asyncio
async def test_async_set_heat_temp(
    event_loop: asyncio.AbstractEventLoop, MockConnectedGateway
):
    body = 0
    temp = 88
    result = event_loop.create_future()
    result.set_result(b"")
    with patch(
        "screenlogicpy.requests.heat.ScreenLogicProtocol.await_send_message",
        return_value=result,
    ) as mockRequest:
        gateway = MockConnectedGateway
        assert await gateway.async_set_heat_temp(body, temp)
        await gateway.async_disconnect()
        assert mockRequest.call_args.args[0] == 12528
        assert mockRequest.call_args.args[1] == struct.pack("<III", 0, body, temp)


@pytest.mark.asyncio
async def test_async_set_heat_mode(
    event_loop: asyncio.AbstractEventLoop, MockConnectedGateway
):
    body = 0
    mode = 3
    result = event_loop.create_future()
    result.set_result(b"")
    with patch(
        "screenlogicpy.requests.heat.ScreenLogicProtocol.await_send_message",
        return_value=result,
    ) as mockRequest:
        gateway = MockConnectedGateway
        assert await gateway.async_set_heat_mode(body, mode)
        await gateway.async_disconnect()
        assert mockRequest.call_args.args[0] == 12538
        assert mockRequest.call_args.args[1] == struct.pack("<III", 0, body, mode)


@pytest.mark.asyncio
async def test_async_set_color_lights(
    event_loop: asyncio.AbstractEventLoop, MockConnectedGateway
):
    mode = 7
    result = event_loop.create_future()
    result.set_result(b"")
    with patch(
        "screenlogicpy.requests.lights.ScreenLogicProtocol.await_send_message",
        return_value=result,
    ) as mockRequest:
        gateway = MockConnectedGateway
        assert await gateway.async_set_color_lights(mode)
        await gateway.async_disconnect()
        assert mockRequest.call_args.args[0] == 12556
        assert mockRequest.call_args.args[1] == struct.pack("<II", 0, mode)


@pytest.mark.asyncio
async def test_async_set_scg_config(
    event_loop: asyncio.AbstractEventLoop, MockConnectedGateway
):
    pool_output = 50
    spa_output = 0
    result = event_loop.create_future()
    result.set_result(b"")
    with patch(
        "screenlogicpy.requests.heat.ScreenLogicProtocol.await_send_message",
        return_value=result,
    ) as mockRequest:
        gateway = MockConnectedGateway
        assert await gateway.async_set_scg_config(pool_output, spa_output)
        await gateway.async_disconnect()
        assert mockRequest.call_args.args[0] == 12576
        assert mockRequest.call_args.args[1] == struct.pack(
            "<IIIII", 0, pool_output, spa_output, 0, 0
        )



from .const_data import (
    EXPECTED_CONFIG_DATA,
    EXPECTED_STATUS_DATA,
    EXPECTED_PUMP_DATA,
    EXPECTED_CHEMISTRY_DATA,
    EXPECTED_SCG_DATA,
    FAKE_CONFIG_RESPONSE,
    FAKE_STATUS_RESPONSE,
    FAKE_PUMP_RESPONSE,
    FAKE_CHEMISTRY_RESPONSE,
    FAKE_SCG_RESPONSE,
)
from screenlogicpy.requests.config import decode_pool_config
from screenlogicpy.requests.status import decode_pool_status
from screenlogicpy.requests.pump import decode_pump_status
from screenlogicpy.requests.chemistry import decode_chemistry
from screenlogicpy.requests.scg import decode_scg_config
from screenlogicpy.requests.utility import makeMessage, takeMessages


def test_decode_config():
    data = {}
    decode_pool_config(FAKE_CONFIG_RESPONSE, data)

    assert data == EXPECTED_CONFIG_DATA


def test_decode_status():
    data = {}
    decode_pool_status(FAKE_STATUS_RESPONSE, data)

    assert data == EXPECTED_STATUS_DATA


def test_decode_pump():
    data = {}
    decode_pump_status(FAKE_PUMP_RESPONSE, data, 0)

    assert data == EXPECTED_PUMP_DATA


def test_decode_chemistry():
    data = {}
    decode_chemistry(FAKE_CHEMISTRY_RESPONSE, data)

    assert data == EXPECTED_CHEMISTRY_DATA


def test_decode_scg():
    data = {}
    decode_scg_config(FAKE_SCG_RESPONSE, data)

    assert data == EXPECTED_SCG_DATA


def test_takeMessages():
    mID1 = 27
    mCD1 = 12593
    mDT1 = FAKE_CHEMISTRY_RESPONSE
    mID2 = 28
    mCD2 = 12573
    mDT2 = FAKE_SCG_RESPONSE

    joined = makeMessage(mID1, mCD1, mDT1) + makeMessage(mID2, mCD2, mDT2)
    messages = takeMessages(joined)

    assert messages[0][0] == mID1
    assert messages[0][1] == mCD1
    assert messages[0][2] == mDT1

    assert messages[1][0] == mID2
    assert messages[1][1] == mCD2
    assert messages[1][2] == mDT2



import struct

from ..const import CODE, CLIENT_ID
from .protocol import ScreenLogicProtocol
from .request import async_make_request


async def async_request_add_client(
    protocol: ScreenLogicProtocol, clientID: int = CLIENT_ID
) -> bool:
    return (
        await async_make_request(
            protocol, CODE.ADD_CLIENT_QUERY, struct.pack("<II", 0, clientID)
        )
        == b""
    )


async def async_request_remove_client(
    protocol: ScreenLogicProtocol, clientID: int = CLIENT_ID
) -> bool:
    return (
        await async_make_request(
            protocol, CODE.REMOVE_CLIENT_QUERY, struct.pack("<II", 0, clientID)
        )
        == b""
    )



from ..const import CODE
from .protocol import ScreenLogicProtocol
from .request import async_make_request


async def async_request_ping(protocol: ScreenLogicProtocol) -> bytes:
    return await async_make_request(protocol, CODE.PING_QUERY) == b""



import struct

from ..const import CODE
from .protocol import ScreenLogicProtocol
from .request import async_make_request
from .utility import getSome, getArray


async def async_request_equipment_config(
    protocol: ScreenLogicProtocol, data: dict
) -> bytes:
    if result := await async_make_request(
        protocol, CODE.EQUIPMENT_QUERY, struct.pack("<2I", 0, 0)
    ):
        decode_equipment_config(result, data)
        return result


def decode_equipment_config(buff: bytes, data: dict) -> None:
    equip = data.setdefault("equipment", {})

    equip["controllerType"], offset = getSome("B", buff, 0)
    equip["hardwareType"], offset = getSome("B", buff, offset)

    equip[f"unknown_at_offset_{offset:02}"], offset = getSome("B", buff, offset)
    equip[f"unknown_at_offset_{offset:02}"], offset = getSome("B", buff, offset)
    equip["controllerData"], offset = getSome("I", buff, offset)

    equip["versionDataArray"], offset = getArray(buff, offset)
    equip["speedDataArray"], offset = getArray(buff, offset)
    equip["valveDataArray"], offset = getArray(buff, offset)
    equip["remoteDataArray"], offset = getArray(buff, offset)
    equip["sensorDataArray"], offset = getArray(buff, offset)
    equip["delayDataArray"], offset = getArray(buff, offset)
    equip["macrosDataArray"], offset = getArray(buff, offset)
    equip["miscDataArray"], offset = getArray(buff, offset)
    equip["lightDataArray"], offset = getArray(buff, offset)
    equip["flowsDataArray"], offset = getArray(buff, offset)
    equip["sgsDataArray"], offset = getArray(buff, offset)
    equip["spaFlowsDataArray"], offset = getArray(buff, offset)



import asyncio

from ..const import MESSAGE, ScreenLogicWarning
from .protocol import ScreenLogicProtocol
from .utility import asyncio_timeout


async def async_make_request(
    protocol: ScreenLogicProtocol, messageCode: int, message: bytes = b""
) -> bytes:
    request = protocol.await_send_message(messageCode, message)
    try:
        async with asyncio_timeout(MESSAGE.COM_TIMEOUT):
            await request
    except asyncio.TimeoutError:
        raise ScreenLogicWarning(
            f"Timeout waiting for response to message code '{messageCode}'"
        )
    if not request.cancelled():
        return request.result()



import asyncio
import struct
import pytest
from unittest.mock import patch

from screenlogicpy import ScreenLogicGateway
from screenlogicpy.client import ClientManager
from screenlogicpy.const import CLIENT_ID, CODE
from .const_data import (
    EXPECTED_STATUS_DATA,
    EXPECTED_CHEMISTRY_DATA,
    FAKE_CONNECT_INFO,
    FAKE_CHEMISTRY_RESPONSE,
    FAKE_STATUS_RESPONSE,
)


@pytest.mark.asyncio()
async def test_sub_unsub(event_loop, MockProtocolAdapter):
    async with MockProtocolAdapter:
        gateway = ScreenLogicGateway()
        code = CODE.STATUS_CHANGED
        clientID = CLIENT_ID

        def callback():
            pass

        bad_unsub = await gateway.async_subscribe_client(callback, code)

        assert not bad_unsub

        await gateway.async_connect(**FAKE_CONNECT_INFO)

        result = event_loop.create_future()
        result.set_result(b"")
        with patch(
            "screenlogicpy.requests.client.ScreenLogicProtocol.await_send_message",
            return_value=result,
        ) as mockSubRequest:
            unsub = await gateway.async_subscribe_client(callback, code)

            assert callable(unsub)
            assert gateway._client_manager._listeners == {
                code: {
                    callback,
                },
            }
            assert gateway._protocol._callbacks == {
                code: (
                    gateway._client_manager._async_common_callback,
                    (code, gateway._data),
                ),
            }
            assert mockSubRequest.call_args.args[0] == 12522
            assert mockSubRequest.call_args.args[1] == struct.pack("<II", 0, clientID)

        with patch(
            "screenlogicpy.requests.client.ScreenLogicProtocol.await_send_message",
            return_value=result,
        ) as mockUnsubRequest:
            unsub()

            assert gateway._client_manager._listeners == {}
            assert gateway._protocol._callbacks == {}

            await asyncio.sleep(0)

            assert mockUnsubRequest.call_args.args[0] == 12524
            assert mockUnsubRequest.call_args.args[1] == struct.pack("<II", 0, clientID)


@pytest.mark.asyncio()
async def test_notify():
    code1 = CODE.STATUS_CHANGED
    code2 = CODE.CHEMISTRY_CHANGED

    cb1_hit = False
    cb2_hit = False
    cb3_hit = False

    def callback1():
        nonlocal cb1_hit
        cb1_hit = True

    def callback2():
        nonlocal cb2_hit
        cb2_hit = True

    def callback3():
        nonlocal cb3_hit
        cb3_hit = True

    cm = ClientManager()
    cm._listeners = {
        code1: {
            callback1,
            callback2,
        },
        code2: {
            callback3,
        },
    }

    status_data = {}
    await cm._async_common_callback(
        FAKE_STATUS_RESPONSE, CODE.STATUS_CHANGED, status_data
    )

    assert cb1_hit
    assert cb2_hit
    assert not cb3_hit

    chem_data = {}
    await cm._async_common_callback(
        FAKE_CHEMISTRY_RESPONSE, CODE.CHEMISTRY_CHANGED, chem_data
    )

    assert cb3_hit

    assert status_data == EXPECTED_STATUS_DATA
    assert chem_data == EXPECTED_CHEMISTRY_DATA


@pytest.mark.asyncio()
async def test_attach_existing(MockProtocolAdapter):
    gateway = ScreenLogicGateway()
    code1 = CODE.STATUS_CHANGED
    code2 = CODE.CHEMISTRY_CHANGED

    cb1_hit = False
    cb2_hit = False
    cb3_hit = False

    def callback1():
        nonlocal cb1_hit
        cb1_hit = True

    def callback2():
        nonlocal cb2_hit
        cb2_hit = True

    def callback3():
        nonlocal cb3_hit
        cb3_hit = True

    gateway._client_manager._listeners = {
        code1: {
            callback1,
            callback2,
        },
        code2: {
            callback3,
        },
    }
    async with MockProtocolAdapter:

        await gateway.async_connect(**FAKE_CONNECT_INFO)

        assert gateway._protocol._callbacks == {
            code1: (
                gateway._client_manager._async_common_callback,
                (code1, gateway._data),
            ),
            code2: (
                gateway._client_manager._async_common_callback,
                (code2, gateway._data),
            ),
        }



"""Discovery for screenlogic gateways."""
import asyncio
import ipaddress
import logging
import socket
import struct

from .const import (  # pylint: disable=relative-beyond-top-level
    SL_GATEWAY_IP,
    SL_GATEWAY_NAME,
    SL_GATEWAY_PORT,
    SL_GATEWAY_SUBTYPE,
    SL_GATEWAY_TYPE,
    ScreenLogicError,
)

DISCOVERY_PAYLOAD = struct.pack("<bbbbbbbb", 1, 0, 0, 0, 0, 0, 0, 0)
DISCOVERY_RESPONSE_FORMAT = "<I4BH2B"
DISCOVERY_RESPONSE_SIZE = struct.calcsize(DISCOVERY_RESPONSE_FORMAT)
DISCOVERY_ADDRESS = "255.255.255.255"
DISCOVERY_PORT = 1444
DISCOVERY_CHKSUM = 2
DISCOVERY_TIMEOUT = 1

_LOGGER = logging.getLogger(__name__)


def create_broadcast_socket():
    """Create a broadcast soscket for discovery."""
    addressfamily = socket.AF_INET
    udp_sock = socket.socket(addressfamily, socket.SOCK_DGRAM)
    udp_sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
    udp_sock.bind(("", 0))
    return udp_sock


def process_discovery_response(data):
    """Process a discovery response."""
    try:
        paddedfmt = (
            DISCOVERY_RESPONSE_FORMAT + str(len(data) - DISCOVERY_RESPONSE_SIZE) + "s"
        )
        (
            chk,
            ip1,
            ip2,
            ip3,
            ip4,
            gateway_port,
            gateway_type,
            gateway_subtype,
            gateway_name,
        ) = struct.unpack(paddedfmt, data)

        # not sure we need to check if "chk" isn't what we wanted.
        if chk != DISCOVERY_CHKSUM:
            raise ScreenLogicError(
                "ScreenLogic Discovery: Unexpected response checksum."
            )

        # make sure we got a good IP address
        received_ip = f"{str(ip1)}.{str(ip2)}.{str(ip3)}.{str(ip4)}"
        gateway_ip = str(ipaddress.ip_address(received_ip))
    except ValueError as err:
        raise ScreenLogicError(
            "ScreenLogic Discovery: Got an invalid IP address from the gateway."
        ) from err
    except NameError as err:
        raise ScreenLogicError(
            "ScreenLogic Discovery: Received garbage from the gateway."
        ) from err
    except Exception as err:
        raise ScreenLogicError(
            "ScreenLogic Discovery: Couldn't get an IP address for the gateway."
        ) from err

    return {
        SL_GATEWAY_IP: gateway_ip,
        SL_GATEWAY_PORT: gateway_port,
        SL_GATEWAY_TYPE: gateway_type,
        SL_GATEWAY_SUBTYPE: gateway_subtype,
        SL_GATEWAY_NAME: gateway_name.decode("utf-8").strip("\0"),
    }


class ScreenLogicDiscoveryProtocol:
    """Implement ScreenLogic discovery protocol."""

    def __init__(self):
        """Init protocol."""
        self.transport = None
        self.hosts = []

    def connection_lost(self, _):
        """Connection lost."""

    def connection_made(self, transport):
        """Connection made."""
        self.transport = transport

    def datagram_received(self, data, _):
        """Response recieved."""
        try:
            self.hosts.append(process_discovery_response(data))
        except Exception as ex:  # pylint: disable=broad-except
            _LOGGER.warning(ex)

    def error_received(self, exc):
        """Error received."""


async def async_discover():
    """Discover screenlogic gateways."""
    loop = asyncio.get_event_loop()
    udp_sock = create_broadcast_socket()
    transport, protocol = await loop.create_datagram_endpoint(
        lambda: ScreenLogicDiscoveryProtocol(),  # pylint: disable=unnecessary-lambda
        sock=udp_sock,
    )
    try:
        transport.sendto(DISCOVERY_PAYLOAD, (DISCOVERY_ADDRESS, DISCOVERY_PORT))
        await asyncio.sleep(DISCOVERY_TIMEOUT)
    finally:
        transport.close()
    return protocol.hosts



import struct
import sys
from typing import List, Tuple

from ..const import CODE, DATA, MESSAGE, UNIT, ScreenLogicError

if sys.version_info[:2] < (3, 11):
    from async_timeout import timeout as asyncio_timeout
else:
    from asyncio import timeout as asyncio_timeout


def makeMessage(msgID: int, msgCode: int, messageData: bytes = b""):
    """Returns packed bytes formatted as a ready-to-send ScreenLogic message."""
    return struct.pack(
        MESSAGE.HEADER_FORMAT + str(len(messageData)) + "s",
        msgID,
        msgCode,
        len(messageData),
        messageData,
    )


def takeMessage(data: bytes) -> Tuple[int, int, bytes]:
    """Return (messageID, messageCode, message) from raw ScreenLogic message bytes."""
    messageBytes = len(data) - MESSAGE.HEADER_LENGTH
    msgID, msgCode, msgLen, message = struct.unpack(
        MESSAGE.HEADER_FORMAT + str(messageBytes) + "s", data
    )
    if msgLen != messageBytes:
        raise ScreenLogicError(
            f"Response length invalid. Claimed: {msgLen}. Received: {messageBytes}. Message ID: {msgID}. Message Code: {msgCode}. Data: {data}"
        )
    if msgCode == CODE.UNKNOWN_ANSWER:
        raise ScreenLogicError("Request rejected")
    return msgID, msgCode, message  # return raw data


def takeMessages(data: bytes) -> List[Tuple[int, int, bytes]]:
    messages = []
    pos = 0
    try:
        while pos < len(data):
            msgID, pos = getSome("H", data, pos)
            msgCode, pos = getSome("H", data, pos)
            msgLength, pos = getSome("I", data, pos)
            message, pos = getSome(f"{msgLength}s", data, pos)
            messages.append([msgID, msgCode, message])
        return messages
    except struct.error as err:
        raise ScreenLogicError(
            f"Unexpected amount of data received. Data: {data}"
        ) from err


def encodeMessageString(string):
    data = string.encode()
    length = len(data)
    over = length % 4
    pad = (4 - over) if over > 0 else 0  # pad string to multiple of 4
    fmt = "<I" + str(length + pad) + "s"
    return struct.pack(fmt, length, data)


def decodeMessageString(data):
    size = struct.unpack_from("<I", data, 0)[0]
    return struct.unpack_from("<" + str(size) + "s", data, struct.calcsize("<I"))[
        0
    ].decode("utf-8")


def getSome(want, buff, offset):
    fmt = want if want.startswith(">") else "<" + want
    newoffset = offset + struct.calcsize(fmt)
    return struct.unpack_from(fmt, buff, offset)[0], newoffset


def getValueAt(buff, offset, want, **kwargs):
    fmt = want if want.startswith(">") else "<" + want
    val = kwargs.get("adjustment", lambda x: x)(
        struct.unpack_from(fmt, buff, offset)[0]
    )
    if name := kwargs.get("name"):
        data = {
            "name": name,
            "value": val,
        }
        if unit := kwargs.get("unit"):
            data["unit"] = unit
        if device_type := kwargs.get("device_type"):
            data["device_type"] = device_type
    else:
        data = val
    newoffset = offset + struct.calcsize(fmt)
    return data, newoffset


def getString(buff, offset):
    fmtLen = "<I"
    offsetLen = offset + struct.calcsize(fmtLen)
    sLen = struct.unpack_from(fmtLen, buff, offset)[0]
    if sLen % 4 != 0:
        sLen += 4 - sLen % 4

    fmt = "<{}{}".format(sLen, "s")
    newoffset = offsetLen + struct.calcsize(fmt)
    return struct.unpack_from(fmt, buff, offsetLen)[0], newoffset


def getArray(buff, offset):
    itemCount, offset = getSome("I", buff, offset)
    items = [0 for x in range(itemCount)]
    for i in range(itemCount):
        items[i], offset = getSome("B", buff, offset)
    offsetPad = 0
    if itemCount % 4 != 0:
        offsetPad = (4 - itemCount % 4) % 4
        offset += offsetPad
    return items, offset


def getTemperatureUnit(data: dict):
    return (
        UNIT.CELSIUS
        if DATA.KEY_CONFIG in data
        and "is_celsius" in data[DATA.KEY_CONFIG]
        and data[DATA.KEY_CONFIG]["is_celsius"]["value"]
        else UNIT.FAHRENHEIT
    )



# flake8: noqa F401
from .button import async_request_pool_button_press
from .config import async_request_pool_config
from .equipment import async_request_equipment_config
from .gateway import async_request_gateway_version
from .heat import async_request_set_heat_mode, async_request_set_heat_setpoint
from .lights import async_request_pool_lights_command
from .login import async_connect_to_gateway
from .ping import async_request_ping
from .pump import async_request_pump_status
from .status import async_request_pool_status
from .chemistry import async_request_chemistry, async_request_set_chem_data
from .scg import async_request_scg_config, async_request_set_scg_config
from .request import async_make_request



import json
from unittest.mock import patch

import pytest
from screenlogicpy.cli import cli
from .const_data import (
    FAKE_GATEWAY_ADDRESS,
    FAKE_GATEWAY_NAME,
    FAKE_GATEWAY_PORT,
    FAKE_CONNECT_INFO,
    EXPECTED_COMPLETE_DATA,
    EXPECTED_DASHBOARD,
    EXPECTED_VERBOSE_PREAMBLE,
)


async def run_cli_test(
    protocol_adapter, system_out, arguments, return_code, expected_output
):
    async with protocol_adapter:
        with patch(
            "screenlogicpy.cli.async_discover", return_value=[FAKE_CONNECT_INFO]
        ):
            assert await cli(arguments.split()) == return_code
            assert system_out.readouterr().out.strip() == expected_output


@pytest.mark.asyncio
async def test_dashboard(MockProtocolAdapter, capsys):
    await run_cli_test(MockProtocolAdapter, capsys, "", 0, EXPECTED_DASHBOARD)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "args, ret, expected",
    [
        (
            "discover",
            0,
            f"{FAKE_GATEWAY_ADDRESS}:{FAKE_GATEWAY_PORT} '{FAKE_GATEWAY_NAME}'",
        ),
        (
            "-v discover",
            0,
            f"Discovered:\n'{FAKE_GATEWAY_NAME}' at {FAKE_GATEWAY_ADDRESS}:{FAKE_GATEWAY_PORT}",
        ),
    ],
)
async def test_discover(MockProtocolAdapter, capsys, args, ret, expected):
    await run_cli_test(MockProtocolAdapter, capsys, args, ret, expected)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "args, ret, expected",
    [
        ("get circuit 500", 0, "0"),
        ("get c 502", 0, "1"),
        ("get c 300", 4, "Invalid circuit number: 300"),
        (
            "-v get c 502",
            0,
            EXPECTED_VERBOSE_PREAMBLE + "Pool Light: On",
        ),
        ("set circuit 500 off", 0, "0"),
        (
            "-v set circuit 502 1",
            0,
            EXPECTED_VERBOSE_PREAMBLE + "Pool Light: On",
        ),
        ("set circuit 900 1", 4, "Invalid circuit number: 900"),
    ],
)
async def test_circuit(MockProtocolAdapter, capsys, args, ret, expected):
    await run_cli_test(MockProtocolAdapter, capsys, args, ret, expected)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "args, ret, expected",
    [
        ("get current-temp 0", 0, "56"),
        ("get t spa", 0, "97"),
    ],
)
async def test_get_current_temp(MockProtocolAdapter, capsys, args, ret, expected):
    await run_cli_test(MockProtocolAdapter, capsys, args, ret, expected)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "args, ret, expected",
    [
        ("get heat-mode pool", 0, "0"),
        ("get hm 1", 0, "3"),
        ("set heat-mode pool 0", 0, "0"),
        (
            "-v set hm 1 heater",
            0,
            EXPECTED_VERBOSE_PREAMBLE + "Spa Heat Mode: Heater",
        ),
    ],
)
async def test_heat_mode(MockProtocolAdapter, capsys, args, ret, expected):
    await run_cli_test(MockProtocolAdapter, capsys, args, ret, expected)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "args, ret, expected",
    [
        ("get heat-temp 0", 0, "86"),
        ("get ht spa", 0, "97"),
        ("set heat-temp pool 86", 0, "86"),
        (
            "-v set ht 1 97",
            0,
            EXPECTED_VERBOSE_PREAMBLE + "Spa Heat Set Point: 97",
        ),
    ],
)
async def test_heat_temp(MockProtocolAdapter, capsys, args, ret, expected):
    await run_cli_test(MockProtocolAdapter, capsys, args, ret, expected)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "args, ret, expected",
    [
        ("get heat-state 0", 0, "0"),
        ("get hs spa", 0, "0"),
    ],
)
async def test_get_heat_state(MockProtocolAdapter, capsys, args, ret, expected):
    await run_cli_test(MockProtocolAdapter, capsys, args, ret, expected)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "args, ret, expected",
    [
        ("get json", 0, json.dumps(EXPECTED_COMPLETE_DATA, indent=2)),
    ],
)
async def test_get_json(MockProtocolAdapter, capsys, args, ret, expected):
    await run_cli_test(MockProtocolAdapter, capsys, args, ret, expected)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "args, ret, expected",
    [
        ("set color-lights romance", 0, "6"),
        (
            "-v set cl 0",
            0,
            EXPECTED_VERBOSE_PREAMBLE + "Set color mode to All Off",
        ),
    ],
)
async def test_set_color_lights(MockProtocolAdapter, capsys, args, ret, expected):
    await run_cli_test(MockProtocolAdapter, capsys, args, ret, expected)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "args, ret, expected",
    [
        ("set salt-generator 100 20", 0, "50 0"),
        (
            "-v set scg 20 0",
            0,
            EXPECTED_VERBOSE_PREAMBLE + "Pool SCG Level: 50 Spa SCG Level: 0",
        ),
        ("set scg * *", 65, "No new SCG values. Nothing to do."),
        ("set scg f *", 66, "Invalid SCG value"),
    ],
)
async def test_set_scg(MockProtocolAdapter, capsys, args, ret, expected):
    await run_cli_test(MockProtocolAdapter, capsys, args, ret, expected)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "args, ret, expected",
    [
        ("set chem-data 7.5 700", 0, "7.5 700"),
        (
            "-v set ch 7.6 650",
            0,
            EXPECTED_VERBOSE_PREAMBLE + "pH Setpoint: 7.5 ORP Setpoint: 700",
        ),
        ("set ch * *", 129, "No new setpoint values. Nothing to do."),
        ("set ch f *", 130, "Invalid Chemistry Setpoint value"),
    ],
)
async def test_set_chemistry(MockProtocolAdapter, capsys, args, ret, expected):
    await run_cli_test(MockProtocolAdapter, capsys, args, ret, expected)



"""Describes a ScreenLogicGateway class for interacting with a Pentair ScreenLogic system."""
import asyncio
import logging
from typing import Awaitable, Callable

from .client import ClientManager
from .const import (
    BODY_TYPE,
    CHEMISTRY,
    CIRCUIT_FUNCTION,
    DATA,
    RANGE,
    SCG,
    ScreenLogicError,
    ScreenLogicWarning,
)
from .requests import (
    async_connect_to_gateway,
    async_request_gateway_version,
    async_request_pool_button_press,
    async_request_pool_config,
    async_request_pool_lights_command,
    async_request_pool_status,
    async_request_pump_status,
    async_request_set_heat_mode,
    async_request_set_heat_setpoint,
    async_request_chemistry,
    async_request_scg_config,
    async_request_set_scg_config,
    async_request_set_chem_data,
    async_make_request,
)
from .requests.protocol import ScreenLogicProtocol


_LOGGER = logging.getLogger(__name__)


class ScreenLogicGateway:
    """Class for interacting and communicating with a ScreenLogic protocol adapter."""

    def __init__(
        self, ip=None, port=80, gtype=0, gsubtype=0, name="Unnamed-Screenlogic-Gateway"
    ):
        self._ip = ip
        self._port = port
        self._type = gtype
        self._subtype = gsubtype
        self._name = name
        self._mac = ""
        self._version = ""
        self._transport: asyncio.Transport = None
        self._protocol: ScreenLogicProtocol = None
        self._is_client = False
        self._data = {}
        self._last = {}
        self._client_manager = ClientManager()

    @property
    def ip(self) -> str:
        return self._ip

    @property
    def port(self) -> int:
        return self._port

    @property
    def name(self) -> str:
        return self._name

    @property
    def mac(self) -> str:
        return self._mac

    @property
    def version(self) -> str:
        return self._version

    @property
    def is_connected(self) -> bool:
        return self._protocol._connected if self._protocol else False

    @property
    def is_client(self) -> bool:
        return self._client_manager.is_client

    async def async_connect(
        self,
        ip=None,
        port=None,
        gtype=None,
        gsubtype=None,
        name=None,
        connection_closed_callback: Callable = None,
    ) -> bool:
        """Connect to the ScreenLogic protocol adapter"""
        if self.is_connected:
            return True

        self._ip = ip if ip else self._ip
        self._port = port if port else self._port
        self._type = gtype if gtype else self._type
        self._subtype = gsubtype if gsubtype else self._subtype
        self._name = name if name else self._name
        self._custom_connection_closed_callback = connection_closed_callback

        if not self._ip:
            raise ScreenLogicError("IP address never provided for connection.")

        _LOGGER.debug("Beginning connection and login sequence")
        connectPkg = await async_connect_to_gateway(
            self._ip, self._port, self._common_connection_closed_callback
        )
        if connectPkg:
            self._transport, self._protocol, self._mac = connectPkg
            self._version = await async_request_gateway_version(self._protocol)
            if self._version:
                _LOGGER.debug("Login successful")
                await self.async_get_config()
                await self._client_manager.attach(self._protocol, self.get_data())
                return True
        _LOGGER.debug("Login failed")
        return False

    async def async_disconnect(self, force=False):
        """Disconnect from the ScreenLogic protocol adapter"""
        if self.is_client:
            await self._client_manager.async_unsubscribe_gateway()

        if not force:
            while self._protocol.requests_pending():
                await asyncio.sleep(1)

        if self._transport and not self._transport.is_closing():
            self._transport.close()

    async def async_update(self) -> bool:
        """
        Update all ScreenLogic data.

        Try to reconnect to the ScreenLogic protocol adapter if needed.
        """
        if not await self.async_connect() or not self._data:
            return False

        _LOGGER.debug("Beginning update of all data")
        await self.async_get_status()
        await self.async_get_pumps()
        await self.async_get_chemistry()
        await self.async_get_scg()
        _LOGGER.debug("Update complete")
        return True

    def get_data(self) -> dict:
        """Return the data."""
        return self._data

    def get_debug(self) -> dict:
        """Return the debug last-received data."""
        return self._last

    async def async_set_circuit(self, circuitID: int, circuitState: int):
        """Set the circuit state for the specified circuit."""
        if not self._is_valid_circuit(circuitID):
            raise ValueError(f"Invalid circuitID: {circuitID}")
        if not self._is_valid_circuit_state(circuitState):
            raise ValueError(f"Invalid circuitState: {circuitState}")

        if await self.async_connect():
            if await async_request_pool_button_press(
                self._protocol, circuitID, circuitState
            ):
                return True
        return False

    async def async_set_heat_temp(self, body: int, temp: int):
        """Set the target temperature for the specified body."""
        if not self._is_valid_body(body):
            raise ValueError(f"Invalid body: {body}")
        if not self._is_valid_heattemp(body, temp):
            raise ValueError(f"Invalid temp ({temp}) for body ({body})")

        if await self.async_connect():
            if await async_request_set_heat_setpoint(self._protocol, body, temp):
                return True
        return False

    async def async_set_heat_mode(self, body: int, mode: int):
        """Set the heating mode for the specified body."""
        if not self._is_valid_body(body):
            raise ValueError(f"Invalid body: {body}")
        if not self._is_valid_heatmode(mode):
            raise ValueError(f"Invalid mode: {mode}")

        if await self.async_connect():
            if await async_request_set_heat_mode(self._protocol, body, mode):
                return True
        return False

    async def async_set_color_lights(self, light_command: int):
        """Set the light show mode for all capable lights."""
        if not self._is_valid_color_mode(light_command):
            raise ValueError(f"Invalid light_command: {light_command}")

        if await self.async_connect():
            if await async_request_pool_lights_command(self._protocol, light_command):
                return True
        return False

    async def async_set_scg_config(self, pool_output: int, spa_output: int):
        """Set the salt-chlorine-generator output for both pool and spa."""
        if not self._is_valid_scg_value(pool_output, BODY_TYPE.POOL):
            raise ValueError(f"Invalid pool_output: {pool_output}")
        if not self._is_valid_scg_value(spa_output, BODY_TYPE.SPA):
            raise ValueError(f"Invalid spa_output: {spa_output}")

        if await self.async_connect():
            if await async_request_set_scg_config(
                self._protocol, pool_output, spa_output
            ):
                return True
        return False

    async def async_set_chem_data(
        self,
        ph_setpoint: float,
        orp_setpoint: int,
        calcium: int,
        alkalinity: int,
        cyanuric: int,
        salt: int,
    ):
        """Set configurable chemistry values."""
        if self._is_valid_ph_setpoint(ph_setpoint):
            ph_setpoint = int(ph_setpoint * 100)
        else:
            raise ValueError(f"Invalid PH Set point: {ph_setpoint}")
        if not self._is_valid_orp_setpoint(orp_setpoint):
            raise ValueError(f"Invalid ORP Set point: {orp_setpoint}")
        if calcium < 0 or alkalinity < 0 or cyanuric < 0 or salt < 0:
            raise ValueError("Invalid Chemistry setting.")

        if await self.async_connect():
            if await async_request_set_chem_data(
                self._protocol,
                ph_setpoint,
                orp_setpoint,
                calcium,
                alkalinity,
                cyanuric,
                salt,
            ):
                return True
        return False

    async def async_subscribe_client(
        self, callback: Callable[..., any], code: int
    ) -> Callable:
        """
        Subscribe client listener to message code.

        Subscribe to push messaging from the ScreenLogic protocol adapter and register a
        callback method to call when a message with the specified message code is received.

        Messages with known codes will be processed to update gateway data before
        callback method is called.
        """
        return await self._client_manager.async_subscribe(callback, code)

    def register_async_message_handler(
        self, message_code: int, handler: Callable[[bytes, any], Awaitable[None]], *argv
    ):
        """
        Register handler for message code.

        Registers an async function to call when a message with the specified message_code is received.
        Only one handler can be registered per message_code. Subsequent registrations will override
        the previous registration.
        """
        if not self._protocol:
            raise ScreenLogicError(
                "Not connected to ScreenLogic gateway. Must connect to gateway before registering handler."
            )
        self._protocol.register_async_message_callback(message_code, handler, *argv)

    def remove_async_message_handler(self, message_code: int):
        """Remove handler for message code."""
        if self._protocol:
            self._protocol.remove_async_message_callback(message_code)

    async def async_send_message(self, message_code: int, message: bytes = b""):
        """Send a message to the ScreenLogic protocol adapter."""
        if not self.is_connected:
            raise ScreenLogicWarning(
                "Not connected to protocol adapter. send_message failed."
            )
        _LOGGER.debug(f"User requesting {message_code}")
        return await async_make_request(self._protocol, message_code, message)

    async def async_get_config(self):
        """Request pool configuration data."""
        if not self.is_connected:
            raise ScreenLogicWarning(
                "Not connected to protocol adapter. get_config failed."
            )
        _LOGGER.debug("Requesting config data")
        self._last[DATA.KEY_CONFIG] = await async_request_pool_config(
            self._protocol, self._data
        )

    async def async_get_status(self):
        """Request pool state data."""
        if not self.is_connected:
            raise ScreenLogicWarning(
                "Not connected to protocol adapter. get_status failed."
            )
        _LOGGER.debug("Requesting pool status")
        self._last["status"] = await async_request_pool_status(
            self._protocol, self._data
        )

    async def async_get_pumps(self):
        """Request all pump state data."""
        if not self.is_connected:
            raise ScreenLogicWarning(
                "Not connected to protocol adapter. get_pumps failed."
            )
        for pumpID in self._data[DATA.KEY_PUMPS]:
            if self._data[DATA.KEY_PUMPS][pumpID]["data"] != 0:
                _LOGGER.debug("Requesting pump %i data", pumpID)
                last_pumps = self._last.setdefault(DATA.KEY_PUMPS, {})
                last_pumps[pumpID] = await async_request_pump_status(
                    self._protocol, self._data, pumpID
                )

    async def async_get_chemistry(self):
        """Request IntelliChem controller data."""
        if not self.is_connected:
            raise ScreenLogicWarning(
                "Not connected to protocol adapter. get_chemistry failed."
            )
        _LOGGER.debug("Requesting chemistry data")
        self._last[DATA.KEY_CHEMISTRY] = await async_request_chemistry(
            self._protocol, self._data
        )

    async def async_get_scg(self):
        """Request salt chlorine generator state data."""
        if not self.is_connected:
            raise ScreenLogicWarning(
                "Not connected to protocol adapter. get_scg failed."
            )
        _LOGGER.debug("Requesting scg data")
        self._last[DATA.KEY_SCG] = await async_request_scg_config(
            self._protocol, self._data
        )

    def _common_connection_closed_callback(self):
        """Perform any needed cleanup."""
        # Future internal cleanup tasks
        if self._custom_connection_closed_callback:
            self._custom_connection_closed_callback()

    def _is_valid_circuit(self, circuit):
        """Validate circuit number."""
        return circuit in self._data[DATA.KEY_CIRCUITS]

    def _is_valid_circuit_state(self, state):
        """Validate circuit state number."""
        return state == 0 or state == 1

    def _is_valid_body(self, body):
        """Validate body of water number."""
        return body in self._data[DATA.KEY_BODIES]

    def _is_valid_heatmode(self, heatmode):
        """Validate heat mode number."""
        return 0 <= heatmode < 5

    def _is_valid_heattemp(self, body, temp):
        """Validate heat tem for body."""
        min_temp = self._data[DATA.KEY_BODIES][int(body)]["min_set_point"]["value"]
        max_temp = self._data[DATA.KEY_BODIES][int(body)]["max_set_point"]["value"]
        return min_temp <= temp <= max_temp

    def _is_valid_color_mode(self, mode):
        """Validate color mode number."""
        return 0 <= mode <= 21

    def _is_valid_scg_value(self, scg_value, body_type):
        """Validate chlorinator value for body."""
        return 0 <= scg_value <= SCG.LIMIT_FOR_BODY[body_type]

    def _is_valid_ph_setpoint(self, ph_setpoint: float):
        """Validate pH setpoint."""
        return (
            CHEMISTRY.RANGE_PH_SETPOINT[RANGE.MIN]
            <= ph_setpoint
            <= CHEMISTRY.RANGE_PH_SETPOINT[RANGE.MAX]
        )

    def _is_valid_orp_setpoint(self, orp_setpoint: int):
        """Validate ORP setpoint."""
        return (
            CHEMISTRY.RANGE_ORP_SETPOINT[RANGE.MIN]
            <= orp_setpoint
            <= CHEMISTRY.RANGE_ORP_SETPOINT[RANGE.MAX]
        )

    # Promote?
    def _has_color_lights(self):
        """Return if any configured lights support color modes."""
        if circuits := self._data.get(DATA.KEY_CIRCUITS, None):
            for circuit in circuits.values():
                if circuit["function"] in CIRCUIT_FUNCTION.GROUP_LIGHTS_COLOR:
                    return True
        return False



import struct

from ..const import CODE
from .protocol import ScreenLogicProtocol
from .request import async_make_request


async def async_request_set_heat_setpoint(
    protocol: ScreenLogicProtocol, body: int, set_point: float
) -> bool:
    return (
        await async_make_request(
            protocol, CODE.SETHEATTEMP_QUERY, struct.pack("<III", 0, body, set_point)
        )
        == b""
    )


async def async_request_set_heat_mode(
    protocol: ScreenLogicProtocol, body: int, heat_mode: int
) -> bool:
    return (
        await async_make_request(
            protocol, CODE.SETHEATMODE_QUERY, struct.pack("<III", 0, body, heat_mode)
        )
        == b""
    )



import pytest
import socket
import struct

from screenlogicpy.discovery import async_discover, process_discovery_response
from .const_data import (
    FAKE_CONNECT_INFO,
    FAKE_GATEWAY_ADDRESS,
    FAKE_GATEWAY_CHK,
    FAKE_GATEWAY_DISCOVERY_PORT,
    FAKE_GATEWAY_NAME,
    FAKE_GATEWAY_PORT,
    FAKE_GATEWAY_TYPE,
    FAKE_GATEWAY_SUB_TYPE,
)
from .fake_gateway import FakeScreenLogicUDPProtocol


@pytest.mark.asyncio
async def test_async_discovery(event_loop):

    _udp_sock = socket.socket(family=socket.AF_INET, type=socket.SOCK_DGRAM)
    _udp_sock.bind(("", FAKE_GATEWAY_DISCOVERY_PORT))

    transport, protocol = await event_loop.create_datagram_endpoint(
        lambda: FakeScreenLogicUDPProtocol(),
        sock=_udp_sock,
    )

    hosts = await async_discover()

    transport.close()

    assert len(hosts) > 0


def test_process_discovery_response():
    ip1, ip2, ip3, ip4 = FAKE_GATEWAY_ADDRESS.split(".")
    response = struct.pack(
        f"<I4BH2B{len(FAKE_GATEWAY_NAME)}s",
        FAKE_GATEWAY_CHK,
        int(ip1),
        int(ip2),
        int(ip3),
        int(ip4),
        FAKE_GATEWAY_PORT,
        FAKE_GATEWAY_TYPE,
        FAKE_GATEWAY_SUB_TYPE,
        bytes(FAKE_GATEWAY_NAME, "UTF-8"),
    )
    data = process_discovery_response(response)
    assert data == FAKE_CONNECT_INFO



import struct

from ..const import CODE, DATA, DEVICE_TYPE, STATE_TYPE, UNIT
from .protocol import ScreenLogicProtocol
from .request import async_make_request
from .utility import getSome


async def async_request_pump_status(
    protocol: ScreenLogicProtocol, data: dict, pumpID: int
) -> bytes:
    if result := await async_make_request(
        protocol, CODE.PUMPSTATUS_QUERY, struct.pack("<II", 0, pumpID)
    ):
        decode_pump_status(result, data, pumpID)
        return result


def decode_pump_status(buff: bytes, data: dict, pumpID: int) -> None:
    pumps: dict = data.setdefault(DATA.KEY_PUMPS, {})

    pump = pumps.setdefault(pumpID, {})

    pump["name"] = ""
    pump["pumpType"], offset = getSome("I", buff, 0)
    pump["state"], offset = getSome("I", buff, offset)

    curW, offset = getSome("I", buff, offset)
    pump["currentWatts"] = {}  # Need to find value when unsupported.

    curR, offset = getSome("I", buff, offset)
    pump["currentRPM"] = {}  # Need to find value when unsupported.

    pump[f"unknown_at_offset_{offset:02}"], offset = getSome("I", buff, offset)

    curG, offset = getSome("I", buff, offset)
    if curG != 255:  # GPM reads 255 when unsupported.
        pump["currentGPM"] = {}

    pump[f"unknown_at_offset_{offset:02}"], offset = getSome("I", buff, offset)

    pump["presets"] = {}
    name = "Default"
    for i in range(8):
        pump["presets"][i] = {}
        pump["presets"][i]["cid"], offset = getSome("I", buff, offset)
        if DATA.KEY_CIRCUITS in data:
            for num, circuit in data[DATA.KEY_CIRCUITS].items():
                if (
                    pump["presets"][i]["cid"] == circuit["device_id"]
                    and name == "Default"
                ):
                    name = circuit["name"]
                    break
        pump["presets"][i]["setPoint"], offset = getSome("I", buff, offset)
        pump["presets"][i]["isRPM"], offset = getSome("I", buff, offset)

    name = name.strip().strip(",") + " Pump"
    pump["name"] = name

    if "currentWatts" in pump:
        pump["currentWatts"] = {
            "name": pump["name"] + " Current Watts",
            "value": curW,
            "unit": UNIT.WATT,
            "device_type": DEVICE_TYPE.POWER,
            "state_type": STATE_TYPE.MEASUREMENT,
        }

    if "currentRPM" in pump:
        pump["currentRPM"] = {
            "name": pump["name"] + " Current RPM",
            "value": curR,
            "unit": UNIT.REVOLUTIONS_PER_MINUTE,
            "state_type": STATE_TYPE.MEASUREMENT,
        }

    if "currentGPM" in pump:
        pump["currentGPM"] = {
            "name": pump["name"] + " Current GPM",
            "value": curG,
            "unit": UNIT.GALLONS_PER_MINUTE,
            "state_type": STATE_TYPE.MEASUREMENT,
        }



# import json
import struct

from ..const import CODE, BODY_TYPE, DATA
from .protocol import ScreenLogicProtocol
from .request import async_make_request
from .utility import getSome, getString


async def async_request_pool_config(protocol: ScreenLogicProtocol, data: dict) -> bytes:
    if result := await async_make_request(
        protocol,
        CODE.CTRLCONFIG_QUERY,
        struct.pack("<2I", 0, 0),  # 0,1 yields different return
    ):
        decode_pool_config(result, data)
        return result


def decode_pool_config(buff: bytes, data: dict) -> dict:
    config: dict = data.setdefault(DATA.KEY_CONFIG, {})

    controllerID, offset = getSome("I", buff, 0)
    config["controller_id"] = {"name": "Controller ID", "value": controllerID}

    bodies: dict = data.setdefault(DATA.KEY_BODIES, {})

    for i in range(2):
        currentBody: dict = bodies.setdefault(i, {})

        minSetPoint, offset = getSome("B", buff, offset)
        MINspName = "{} Minimum Set Point".format(BODY_TYPE.NAME_FOR_NUM[i])
        currentBody["min_set_point"] = {"name": MINspName, "value": minSetPoint}
        maxSetPoint, offset = getSome("B", buff, offset)
        MAXspName = "{} Maximum Set Point".format(BODY_TYPE.NAME_FOR_NUM[i])
        currentBody["max_set_point"] = {"name": MAXspName, "value": maxSetPoint}

    degC, offset = getSome("B", buff, offset)
    config["is_celsius"] = {"name": "Is Celsius", "value": degC}

    controllerType, offset = getSome("B", buff, offset)
    config["controller_type"] = controllerType

    hwType, offset = getSome("B", buff, offset)
    config["hardware_type"] = hwType

    controllerbuff, offset = getSome("B", buff, offset)
    config["controller_buffer"] = controllerbuff

    equipFlags, offset = getSome("I", buff, offset)
    config["equipment_flags"] = equipFlags

    paddedGenName, offset = getString(buff, offset)
    genCircuitName = paddedGenName.decode("utf-8").strip("\0")
    config["generic_circuit_name"] = {
        "name": "Default Circuit Name",
        "value": genCircuitName,
    }

    circuitCount, offset = getSome("I", buff, offset)
    config["circuit_count"] = {"name": "Number of Circuits", "value": circuitCount}

    circuits: dict = data.setdefault(DATA.KEY_CIRCUITS, {})

    for i in range(circuitCount):

        circuitID, offset = getSome("i", buff, offset)

        currentCircuit = circuits.setdefault(circuitID, {})

        currentCircuit["id"] = circuitID

        paddedName, offset = getString(buff, offset)
        circuitName = paddedName.decode("utf-8").strip("\0")
        currentCircuit["name"] = circuitName

        cNameIndex, offset = getSome("B", buff, offset)
        currentCircuit["name_index"] = cNameIndex

        cFunction, offset = getSome("B", buff, offset)
        currentCircuit["function"] = cFunction

        cInterface, offset = getSome("B", buff, offset)
        currentCircuit["interface"] = cInterface

        cFlags, offset = getSome("B", buff, offset)
        currentCircuit["flags"] = cFlags

        cColorSet, offset = getSome("B", buff, offset)
        currentCircuit["color_set"] = cColorSet

        cColorPos, offset = getSome("B", buff, offset)
        currentCircuit["color_position"] = cColorPos

        cColorStagger, offset = getSome("B", buff, offset)
        currentCircuit["color_stagger"] = cColorStagger

        cDeviceID, offset = getSome("B", buff, offset)
        currentCircuit["device_id"] = cDeviceID

        cDefaultRT, offset = getSome("H", buff, offset)
        currentCircuit["default_rt"] = cDefaultRT

        currentCircuit[f"unknown_at_offset_{offset:02}"], offset = getSome(
            "B", buff, offset
        )

        currentCircuit[f"unknown_at_offset_{offset:02}"], offset = getSome(
            "B", buff, offset
        )

    colorCount, offset = getSome("I", buff, offset)
    config["color_count"] = {"name": "Number of Colors", "value": colorCount}

    colors = config.setdefault(DATA.KEY_COLORS, [{} for x in range(colorCount)])

    for i in range(colorCount):
        paddedColorName, offset = getString(buff, offset)
        colorName = paddedColorName.decode("utf-8").strip("\0")
        rgbR, offset = getSome("I", buff, offset)
        rgbG, offset = getSome("I", buff, offset)
        rgbB, offset = getSome("I", buff, offset)
        colors[i] = {"name": colorName, "value": (rgbR, rgbG, rgbB)}

    pumpCircuitCount = 8

    pumps: dict = data.setdefault(DATA.KEY_PUMPS, {})

    for i in range(pumpCircuitCount):
        currentPump = pumps.setdefault(i, {})

        pumpData, offset = getSome("B", buff, offset)
        currentPump["data"] = pumpData

    interfaceTabFlags, offset = getSome("I", buff, offset)
    config["interface_tab_flags"] = interfaceTabFlags

    showAlarms, offset = getSome("I", buff, offset)
    config["show_alarms"] = showAlarms



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="screenlogicpy",
    version="0.7.2",
    author="Kevin Worrel",
    author_email="kevinworrel@yahoo.com",
    description="Interface for Pentair ScreenLogic connected pool controllers over IP via Python",
    license="GPLv3",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/dieselrabbit/screenlogicpy",
    packages=setuptools.find_packages(include=["screenlogicpy", "screenlogicpy.*"]),
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Development Status :: 3 - Alpha",
    ],
    python_requires=">=3.8",
    install_requires=[
        "async_timeout>=3.0.0",
    ],
    entry_points={"console_scripts": ["screenlogicpy=screenlogicpy.__main__:main"]},
)



import struct

from ..const import CODE
from .protocol import ScreenLogicProtocol
from .request import async_make_request


async def async_request_pool_lights_command(
    protocol: ScreenLogicProtocol, light_command: int
) -> bool:
    return (
        await async_make_request(
            protocol, CODE.LIGHTCOMMAND_QUERY, struct.pack("<II", 0, light_command)
        )
        == b""
    )



import asyncio
import string
import json
import argparse
from screenlogicpy.discovery import async_discover
from screenlogicpy.gateway import ScreenLogicGateway
from screenlogicpy.const import (
    BODY_TYPE,
    COLOR_MODE,
    CHEMISTRY,
    DATA,
    EQUIPMENT,
    HEAT_MODE,
    ON_OFF,
    RANGE,
    SL_GATEWAY_IP,
    SL_GATEWAY_NAME,
    SL_GATEWAY_PORT,
    ScreenLogicError,
    ScreenLogicWarning,
)


def cliFormat(name: str):
    table = str.maketrans(" ", "_", string.punctuation)
    return name.translate(table).lower()


def cliFormatDict(mapping: dict):
    return {
        cliFormat(key)
        if isinstance(key, str)
        else key: cliFormat(value)
        if isinstance(value, str)
        else value
        for key, value in mapping.items()
    }


def optionsFromDict(mapping: dict):
    options = []
    for key, value in cliFormatDict(mapping).items():
        options.extend((str(key), str(value)))
    return options


# Entry function
async def cli(cli_args):
    """Handle command line args"""

    def vFormat(slElement: dict, slClass=None):
        if args.verbose:
            if slClass:
                return (
                    f"{slElement['name']}: {slClass.NAME_FOR_NUM[slElement['value']]}"
                )
            else:
                return f"{slElement['name']}: {slElement['value']}"
        else:
            return slElement["value"]

    # Parser functions
    async def async_get_circuit():
        if not int(args.circuit_num) in gateway.get_data()[DATA.KEY_CIRCUITS]:
            print(f"Invalid circuit number: {args.circuit_num}")
            return 4
        print(
            vFormat(
                gateway.get_data()[DATA.KEY_CIRCUITS][int(args.circuit_num)],
                ON_OFF,
            )
        )
        return 0

    async def async_set_circuit():
        state = 0
        if args.state == "1" or args.state.lower() == "on":
            state = 1
        if not int(args.circuit_num) in gateway.get_data()[DATA.KEY_CIRCUITS]:
            print(f"Invalid circuit number: {args.circuit_num}")
            return 4
        if await gateway.async_set_circuit(int(args.circuit_num), state):
            await gateway.async_update()
        else:
            return 4
        print(
            vFormat(
                gateway.get_data()[DATA.KEY_CIRCUITS][int(args.circuit_num)],
                ON_OFF,
            )
        )
        return 0

    async def async_get_heat_mode():
        body = 0
        if args.body == "1" or args.body.lower() == "spa":
            body = 1
        print(
            vFormat(
                gateway.get_data()[DATA.KEY_BODIES][int(body)]["heat_mode"],
                HEAT_MODE,
            )
        )
        return 0

    async def async_set_heat_mode():
        body = 0
        mode = 0
        if args.body == "1" or args.body.lower() == "spa":
            body = 1
        if args.mode in cliFormatDict(HEAT_MODE.NUM_FOR_NAME):
            mode = cliFormatDict(HEAT_MODE.NUM_FOR_NAME)[args.mode]
        else:
            mode = int(args.mode)
        if await gateway.async_set_heat_mode(int(body), mode):
            await gateway.async_update()
        else:
            return 8
        print(
            vFormat(
                gateway.get_data()[DATA.KEY_BODIES][int(body)]["heat_mode"],
                HEAT_MODE,
            )
        )
        return 0

    async def async_get_heat_temp():
        body = 0
        if args.body == "1" or args.body.lower() == "spa":
            body = 1
        print(
            vFormat(
                gateway.get_data()[DATA.KEY_BODIES][int(body)]["heat_set_point"],
            )
        )
        return 0

    async def async_set_heat_temp():
        body = 0
        if args.body == "1" or args.body.lower() == "spa":
            body = 1
        if args.temp != -1:
            if await gateway.async_set_heat_temp(int(body), int(args.temp)):
                await gateway.async_update()
            else:
                return 16
        print(
            vFormat(
                gateway.get_data()[DATA.KEY_BODIES][int(body)]["heat_set_point"],
            )
        )
        return 0

    async def async_get_heat_state():
        body = 0
        if args.body == "1" or args.body.lower() == "spa":
            body = 1
        print(
            vFormat(
                gateway.get_data()[DATA.KEY_BODIES][int(body)]["heat_status"],
                ON_OFF,
            )
        )
        return 0

    async def async_get_current_temp():
        body = 0
        if args.body == "1" or args.body.lower() == "spa":
            body = 1
        print(
            vFormat(
                gateway.get_data()[DATA.KEY_BODIES][int(body)]["last_temperature"],
            )
        )
        return 0

    async def async_set_color_light():
        mode = cliFormatDict(COLOR_MODE.NUM_FOR_NAME).get(args.mode)
        if mode is None:
            mode = int(args.mode)
        if await gateway.async_set_color_lights(mode):
            print(
                f"Set color mode to {COLOR_MODE.NAME_FOR_NUM[mode]}"
                if args.verbose
                else mode
            )
            return 0
        return 32

    async def async_set_scg_config():
        if args.scg_pool == "*" and args.scg_spa == "*":
            print("No new SCG values. Nothing to do.")
            return 65

        scg_data = gateway.get_data()[DATA.KEY_SCG]
        try:
            scg_1 = (
                scg_data["scg_level1"]["value"]
                if args.scg_pool == "*"
                else int(args.scg_pool)
            )
            scg_2 = (
                scg_data["scg_level2"]["value"]
                if args.scg_spa == "*"
                else int(args.scg_spa)
            )
        except ValueError:
            print("Invalid SCG value")
            return 66

        if await gateway.async_set_scg_config(scg_1, scg_2):
            await gateway.async_update()
            new_data = gateway.get_data()
            print(
                vFormat(new_data[DATA.KEY_SCG]["scg_level1"]),
                vFormat(new_data[DATA.KEY_SCG]["scg_level2"]),
            )
            return 0
        return 64

    async def async_set_chem_data():
        if args.ph_setpoint == "*" and args.orp_setpoint == "*":
            print("No new setpoint values. Nothing to do.")
            return 129

        chem_data = gateway.get_data()[DATA.KEY_CHEMISTRY]
        try:
            ph = (
                chem_data["ph_setpoint"]["value"]
                if args.ph_setpoint == "*"
                else float(args.ph_setpoint)
            )
            orp = (
                chem_data["orp_setpoint"]["value"]
                if args.orp_setpoint == "*"
                else int(args.orp_setpoint)
            )
        except ValueError:
            print("Invalid Chemistry Setpoint value")
            return 130

        ch = chem_data["calcium_harness"]["value"]
        ta = chem_data["total_alkalinity"]["value"]
        ca = chem_data["cya"]["value"]
        sa = chem_data["salt_tds_ppm"]["value"]

        if await gateway.async_set_chem_data(ph, orp, ch, ta, ca, sa):
            await asyncio.sleep(3)
            await gateway.async_update()
            new_data = gateway.get_data()
            print(
                vFormat(new_data[DATA.KEY_CHEMISTRY]["ph_setpoint"]),
                vFormat(new_data[DATA.KEY_CHEMISTRY]["orp_setpoint"]),
            )
            return 0
        return 128

    async def async_get_json():
        print(json.dumps(gateway.get_data(), indent=2))
        return 0

    option_parser = argparse.ArgumentParser(
        description="Interface for Pentair Screenlogic gateway"
    )

    option_parser.add_argument("-v", "--verbose", action="store_true")
    option_parser.add_argument("-i", "--ip")
    option_parser.add_argument("-p", "--port", default=80)

    subparsers = option_parser.add_subparsers(dest="action")

    # Discover command
    # pylint: disable=unused-variable
    discover_parser = subparsers.add_parser(  # noqa F841
        "discover", help="Attempt to discover all available ScreenLogic gateways"
    )

    # Get options
    get_parser = subparsers.add_parser("get", help="Gets the option specified")
    get_subparsers = get_parser.add_subparsers(dest="get_option")
    get_subparsers.required = True

    ARGUMENT_CIRCUIT_NUM = {
        "dest": "circuit_num",
        "metavar": "CIRCUIT_NUM",
        "type": int,
        "help": "Circuit number",
    }
    get_circuit_parser = get_subparsers.add_parser("circuit", aliases=["c"])
    get_circuit_parser.add_argument(**ARGUMENT_CIRCUIT_NUM)
    get_circuit_parser.set_defaults(async_func=async_get_circuit)

    body_options = optionsFromDict(BODY_TYPE.NAME_FOR_NUM)
    ARGUMENT_BODY = {
        "dest": "body",
        "metavar": "BODY",
        "type": str,
        "choices": body_options,
        "help": f"Body of water. One of: {body_options}",
    }
    get_heat_mode_parser = get_subparsers.add_parser("heat-mode", aliases=["hm"])
    get_heat_mode_parser.add_argument(**ARGUMENT_BODY)
    get_heat_mode_parser.set_defaults(async_func=async_get_heat_mode)

    get_heat_temp_parser = get_subparsers.add_parser("heat-temp", aliases=["ht"])
    get_heat_temp_parser.add_argument(**ARGUMENT_BODY)
    get_heat_temp_parser.set_defaults(async_func=async_get_heat_temp)

    get_heat_state_parser = get_subparsers.add_parser("heat-state", aliases=["hs"])
    get_heat_state_parser.add_argument(**ARGUMENT_BODY)
    get_heat_state_parser.set_defaults(async_func=async_get_heat_state)

    get_current_temp_parser = get_subparsers.add_parser("current-temp", aliases=["t"])
    get_current_temp_parser.add_argument(**ARGUMENT_BODY)
    get_current_temp_parser.set_defaults(async_func=async_get_current_temp)

    get_json_parser = get_subparsers.add_parser("json", aliases=["j"])
    get_json_parser.set_defaults(async_func=async_get_json)

    # Set options
    set_parser = subparsers.add_parser("set")
    set_subparsers = set_parser.add_subparsers(dest="set_option")
    set_subparsers.required = True

    on_off_options = optionsFromDict(ON_OFF.NAME_FOR_NUM)
    set_circuit_parser = set_subparsers.add_parser("circuit", aliases=["c"])
    set_circuit_parser.add_argument(**ARGUMENT_CIRCUIT_NUM)
    set_circuit_parser.add_argument(
        "state",
        metavar="STATE",
        type=str,
        choices=on_off_options,
        help=f"State to set. One of {on_off_options}",
    )

    cl_options = optionsFromDict(COLOR_MODE.NAME_FOR_NUM)
    set_circuit_parser.set_defaults(async_func=async_set_circuit)
    set_color_light_parser = set_subparsers.add_parser("color-lights", aliases=["cl"])
    set_color_light_parser.add_argument(
        "mode",
        metavar="MODE",
        type=str,
        choices=cl_options,
        help=f"Color lights command, color or show. One of :{cl_options}",
    )
    set_color_light_parser.set_defaults(async_func=async_set_color_light)

    set_heat_mode_parser = set_subparsers.add_parser("heat-mode", aliases=["hm"])
    set_heat_mode_parser.add_argument(**ARGUMENT_BODY)
    hm_options = optionsFromDict(HEAT_MODE.NAME_FOR_NUM)
    set_heat_mode_parser.add_argument(
        "mode",
        metavar="MODE",
        type=str,
        choices=hm_options,
        default=hm_options[0],
        help=f"Heat mode to set. One of: {hm_options}",
    )
    set_heat_mode_parser.set_defaults(async_func=async_set_heat_mode)

    set_heat_temp_parser = set_subparsers.add_parser("heat-temp", aliases=["ht"])
    set_heat_temp_parser.add_argument(**ARGUMENT_BODY)
    set_heat_temp_parser.add_argument(
        "temp",
        type=int,
        metavar="TEMP",
        default=None,
        help="Temperature to set in same unit of measurement as controller settings",
    )
    set_heat_temp_parser.set_defaults(async_func=async_set_heat_temp)

    set_scg_config_parser = set_subparsers.add_parser("salt-generator", aliases=["scg"])
    set_scg_config_parser.add_argument(
        "scg_pool",
        type=str,
        metavar="POOL_PCT",
        default=None,
        help="Chlorinator output for when system is in POOL mode. 0-100, or * to keep current value.",
    )
    set_scg_config_parser.add_argument(
        "scg_spa",
        type=str,
        metavar="SPA_PCT",
        default=None,
        help="Chlorinator output for when system is in SPA mode. 0-20, or * to keep current value.",
    )
    set_scg_config_parser.set_defaults(async_func=async_set_scg_config)

    set_chem_data_parser = set_subparsers.add_parser("chem-data", aliases=["ch"])
    set_chem_data_parser.add_argument(
        "ph_setpoint",
        type=str,
        metavar="PH_SETPOINT",
        default=None,
        help=f"PH set point for IntelliChem. {CHEMISTRY.RANGE_PH_SETPOINT[RANGE.MIN]}-{CHEMISTRY.RANGE_PH_SETPOINT[RANGE.MAX]}, or * to keep current value.",
    )
    set_chem_data_parser.add_argument(
        "orp_setpoint",
        type=str,
        metavar="ORP_SETPOINT",
        default=None,
        help=f"ORP set point for IntelliChem. {CHEMISTRY.RANGE_ORP_SETPOINT[RANGE.MIN]}-{CHEMISTRY.RANGE_ORP_SETPOINT[RANGE.MAX]}, or * to keep current value.",
    )
    set_chem_data_parser.set_defaults(async_func=async_set_chem_data)

    args = option_parser.parse_args(cli_args)

    try:
        host = {SL_GATEWAY_IP: args.ip, SL_GATEWAY_PORT: args.port}
        discovered = False
        if not host[SL_GATEWAY_IP]:
            # Try to discover gateway
            hosts = await async_discover()
            # Host(s) found
            if len(hosts) > 0:
                discovered = True
                if args.action == "discover":
                    if args.verbose:
                        print("Discovered:")
                    for host in hosts:
                        if args.verbose:
                            print(
                                "'{}' at {}:{}".format(
                                    host[SL_GATEWAY_NAME],
                                    host[SL_GATEWAY_IP],
                                    host[SL_GATEWAY_PORT],
                                )
                            )
                        else:
                            print(
                                "{}:{} '{}'".format(
                                    host[SL_GATEWAY_IP],
                                    host[SL_GATEWAY_PORT],
                                    host[SL_GATEWAY_NAME],
                                )
                            )
                    return 0

                # For CLI commands that don't specify an ip address, auto use the first gateway discovered
                # Good for most cases where only one exists on the network
                host = hosts[0]

            else:
                print("No ScreenLogic gateways found.")
                return 1

        gateway = ScreenLogicGateway()

        await gateway.async_connect(**host)

        await gateway.async_update()

        if DATA.KEY_CONFIG not in gateway.get_data():
            return 1

        def print_gateway():
            verb = "Using"
            if discovered:
                verb = "Discovered"
            print(
                "{} '{}' at {}:{}".format(verb, gateway.name, gateway.ip, gateway.port)
            )
            print(
                EQUIPMENT.CONTROLLER_HARDWARE[
                    gateway.get_data()[DATA.KEY_CONFIG]["controller_type"]
                ][gateway.get_data()[DATA.KEY_CONFIG]["hardware_type"]]
            )
            if args.verbose:
                print(f"Version: {gateway.version}")

        def print_circuits():
            print("{}  {}  {}".format("ID".rjust(3), "STATE", "NAME"))
            print("--------------------------")
            for id in gateway.get_data()[DATA.KEY_CIRCUITS]:
                circuit = gateway.get_data()[DATA.KEY_CIRCUITS][int(id)]
                print(
                    "{}  {}  {}".format(
                        circuit["id"],
                        ON_OFF.NAME_FOR_NUM[circuit["value"]].rjust(5),
                        circuit["name"],
                    )
                )

        def print_heat():
            for id in gateway.get_data()[DATA.KEY_BODIES]:
                body = gateway.get_data()[DATA.KEY_BODIES][int(id)]
                print(
                    "{} temperature is last {}{}".format(
                        BODY_TYPE.NAME_FOR_NUM[body["body_type"]["value"]],
                        body["last_temperature"]["value"],
                        body["last_temperature"]["unit"],
                    )
                )
                print(
                    "{}: {}{}".format(
                        body["heat_set_point"]["name"],
                        body["heat_set_point"]["value"],
                        body["last_temperature"]["unit"],
                    )
                )
                print(
                    "{}: {}".format(
                        body["heat_status"]["name"],
                        HEAT_MODE.NAME_FOR_NUM[body["heat_status"]["value"]],
                    )
                )
                print(
                    "{}: {}".format(
                        body["heat_mode"]["name"],
                        HEAT_MODE.NAME_FOR_NUM[body["heat_mode"]["value"]],
                    )
                )
                print("--------------------------")

        def print_dashboard():
            print_gateway()
            print("**************************")
            print_heat()
            print("**************************")
            print_circuits()
            print("**************************")

        if args.action is None:
            print_dashboard()
            await gateway.async_disconnect()
            return 0

        if args.verbose:
            print_gateway()
        result = await args.async_func()
        await gateway.async_disconnect()
        return result

    except (ScreenLogicError, ScreenLogicWarning) as err:
        print(err)
        return 128



# import json
import struct

from ..const import (
    CHEMISTRY,
    CHEM_DOSING_STATE,
    CODE,
    DATA,
    DEVICE_TYPE,
    ON_OFF,
    STATE_TYPE,
    UNIT,
)
from .protocol import ScreenLogicProtocol
from .request import async_make_request
from .utility import getSome, getTemperatureUnit


async def async_request_chemistry(protocol: ScreenLogicProtocol, data: dict) -> bytes:
    if result := await async_make_request(
        protocol, CODE.CHEMISTRY_QUERY, struct.pack("<I", 0)
    ):
        decode_chemistry(result, data)
        return result


# pylint: disable=unused-variable
def decode_chemistry(buff: bytes, data: dict) -> None:
    def is_set(bits, mask) -> bool:
        return True if (bits & mask) == mask else False

    chemistry: dict = data.setdefault(DATA.KEY_CHEMISTRY, {})

    offset = 0

    # size of msg?
    chemistry[f"unknown_at_offset_{offset:02}"], offset = getSome(
        "I", buff, offset
    )  # byte offset 0

    # unknown value
    chemistry[f"unknown_at_offset_{offset:02}"], offset = getSome(
        "B", buff, offset
    )  # byte offset 4

    pH, offset = getSome(">H", buff, offset)  # 5
    chemistry["current_ph"] = {
        "name": "Current pH",
        "value": (pH / 100),
        "unit": UNIT.PH,
        "state_type": STATE_TYPE.MEASUREMENT,
    }

    orp, offset = getSome(">H", buff, offset)  # 7
    chemistry["current_orp"] = {
        "name": "Current ORP",
        "value": orp,
        "unit": UNIT.MILLIVOLT,
        "state_type": STATE_TYPE.MEASUREMENT,
    }

    pHSetpoint, offset = getSome(">H", buff, offset)  # 9
    chemistry["ph_setpoint"] = {
        "name": "pH Setpoint",
        "value": (pHSetpoint / 100),
        "unit": UNIT.PH,
    }

    orpSetpoint, offset = getSome(">H", buff, offset)  # 11
    chemistry["orp_setpoint"] = {
        "name": "ORP Setpoint",
        "value": orpSetpoint,
        "unit": UNIT.MILLIVOLT,
    }

    pHDoseTime, offset = getSome(">I", buff, offset)  # 13
    chemistry["ph_last_dose_time"] = {
        "name": "Last pH Dose Time",
        "value": pHDoseTime,
        "unit": UNIT.SECOND,
        "device_type": DEVICE_TYPE.DURATION,
        "state_type": STATE_TYPE.TOTAL_INCREASING,
    }

    orpDoseTime, offset = getSome(">I", buff, offset)  # 17
    chemistry["orp_last_dose_time"] = {
        "name": "Last ORP Dose Time",
        "value": orpDoseTime,
        "unit": UNIT.SECOND,
        "device_type": DEVICE_TYPE.DURATION,
        "state_type": STATE_TYPE.TOTAL_INCREASING,
    }

    pHDoseVolume, offset = getSome(">H", buff, offset)  # 21
    chemistry["ph_last_dose_volume"] = {
        "name": "Last pH Dose Volume",
        "value": pHDoseVolume,
        "unit": UNIT.MILLILITER,
        "device_type": DEVICE_TYPE.VOLUME,
        "state_type": STATE_TYPE.TOTAL_INCREASING,
    }

    orpDoseVolume, offset = getSome(">H", buff, offset)  # 23
    chemistry["orp_last_dose_volume"] = {
        "name": "Last ORP Dose Volume",
        "value": orpDoseVolume,
        "unit": UNIT.MILLILITER,
        "device_type": DEVICE_TYPE.VOLUME,
        "state_type": STATE_TYPE.TOTAL_INCREASING,
    }

    pHSupplyLevel, offset = getSome("B", buff, offset)  # 25
    chemistry["ph_supply_level"] = {
        "name": "pH Supply Level",
        "value": pHSupplyLevel,
        "state_type": STATE_TYPE.MEASUREMENT,
    }

    orpSupplyLevel, offset = getSome("B", buff, offset)  # 26 (21)
    chemistry["orp_supply_level"] = {
        "name": "ORP Supply Level",
        "value": orpSupplyLevel,
        "state_type": STATE_TYPE.MEASUREMENT,
    }

    saturation, offset = getSome("B", buff, offset)  # 27
    chemistry["saturation"] = {
        "name": "Saturation Index",
        "value": (saturation - 256) / 100
        if is_set(saturation, 0x80)
        else saturation / 100,
        "unit": UNIT.SATURATION_INDEX,
        "state_type": STATE_TYPE.MEASUREMENT,
    }

    cal, offset = getSome(">H", buff, offset)  # 28
    chemistry["calcium_harness"] = {
        "name": "Calcium Hardness",
        "value": cal,
        "unit": UNIT.PARTS_PER_MILLION,
    }

    cya, offset = getSome(">H", buff, offset)  # 30
    chemistry["cya"] = {
        "name": "Cyanuric Acid",
        "value": cya,
        "unit": "ppm",
    }

    alk, offset = getSome(">H", buff, offset)  # 32
    chemistry["total_alkalinity"] = {
        "name": "Total Alkalinity",
        "value": alk,
        "unit": UNIT.PARTS_PER_MILLION,
    }

    saltPPM, offset = getSome("B", buff, offset)  # 34
    chemistry["salt_tds_ppm"] = {
        "name": "Salt/TDS",
        "value": (saltPPM * 50),
        "unit": UNIT.PARTS_PER_MILLION,
    }

    # Probe temp unit is Celsius?
    probIsC, offset = getSome("B", buff, offset)  # 35
    chemistry["probe_is_celsius"] = probIsC

    temperature_unit = getTemperatureUnit(data)

    waterTemp, offset = getSome("B", buff, offset)  # 36
    chemistry["ph_probe_water_temp"] = {
        "name": "pH Probe Water Temperature",
        "value": waterTemp,
        "unit": temperature_unit,
        "device_type": DEVICE_TYPE.TEMPERATURE,
        "state_type": STATE_TYPE.MEASUREMENT,
    }

    alerts = chemistry.setdefault(DATA.KEY_ALERTS, {})

    alarms, offset = getSome("B", buff, offset)  # 37 (32)
    alerts["_raw"] = alarms

    alerts["flow_alarm"] = {
        "name": "Flow Alarm",
        "value": ON_OFF.from_bool(is_set(alarms, CHEMISTRY.FLAG_ALARM_FLOW)),
        "device_type": DEVICE_TYPE.ALARM,
    }
    alerts["ph_alarm"] = {
        "name": "pH Alarm",
        "value": ON_OFF.from_bool(is_set(alarms, CHEMISTRY.FLAG_ALARM_PH)),
        "device_type": DEVICE_TYPE.ALARM,
    }
    alerts["orp_alarm"] = {
        "name": "ORP Alarm",
        "value": ON_OFF.from_bool(is_set(alarms, CHEMISTRY.FLAG_ALARM_ORP)),
        "device_type": DEVICE_TYPE.ALARM,
    }
    alerts["ph_supply_alarm"] = {
        "name": "pH Supply Alarm",
        "value": ON_OFF.from_bool(is_set(alarms, CHEMISTRY.FLAG_ALARM_PH_SUPPLY)),
        "device_type": DEVICE_TYPE.ALARM,
    }
    alerts["orp_supply_alarm"] = {
        "name": "ORP Supply Alarm",
        "value": ON_OFF.from_bool(is_set(alarms, CHEMISTRY.FLAG_ALARM_ORP_SUPPLY)),
        "device_type": DEVICE_TYPE.ALARM,
    }
    alerts["probe_fault_alarm"] = {
        "name": "Probe Fault",
        "value": ON_OFF.from_bool(is_set(alarms, CHEMISTRY.FLAG_ALARM_PROBE_FAULT)),
        "device_type": DEVICE_TYPE.ALARM,
    }

    notifications = chemistry.setdefault(DATA.KEY_NOTIFICATIONS, {})

    warnings, offset = getSome("B", buff, offset)  # 38 (33)
    notifications["_raw"] = warnings

    notifications["ph_lockout"] = {
        "name": "pH Lockout",
        "value": ON_OFF.from_bool(is_set(warnings, CHEMISTRY.FLAG_WARNING_PH_LOCKOUT)),
    }
    notifications["ph_limit"] = {
        "name": "pH Daily Limit Reached",
        "value": ON_OFF.from_bool(is_set(warnings, CHEMISTRY.FLAG_WARNING_PH_LIMIT)),
    }
    notifications["orp_limit"] = {
        "name": "ORP Daily Limit Reached",
        "value": ON_OFF.from_bool(is_set(warnings, CHEMISTRY.FLAG_WARNING_ORP_LIMIT)),
    }

    status, offset = getSome("B", buff, offset)  # 39 (34)
    chemistry["status"] = status
    # notifications["corrosive"] = {
    #    "name": "Corrosive",
    #    "value": ON_OFF.from_bool(is_set(status, CHEMISTRY.FLAG_STATUS_CORROSIVE)),
    # }
    # notifications["scaling"] = {
    #    "name": "Scaling",
    #    "value": ON_OFF.from_bool(is_set(status, CHEMISTRY.FLAG_STATUS_SCALING)),
    # }

    chemistry["ph_dosing_state"] = {
        "name": "pH Dosing State",
        "value": (status & CHEMISTRY.MASK_STATUS_PH_DOSING) >> 4,
        "device_type": DEVICE_TYPE.ENUM,
        "enum_options": [v for v in CHEM_DOSING_STATE.NAME_FOR_NUM.values()],
    }
    chemistry["orp_dosing_state"] = {
        "name": "ORP Dosing State",
        "value": (status & CHEMISTRY.MASK_STATUS_ORP_DOSING) >> 6,
        "device_type": DEVICE_TYPE.ENUM,
        "enum_options": [v for v in CHEM_DOSING_STATE.NAME_FOR_NUM.values()],
    }

    flags, offset = getSome("B", buff, offset)  # 40 (35)
    chemistry["flags"] = flags

    vMinor, offset = getSome("B", buff, offset)  # 41 (36)
    vMajor, offset = getSome("B", buff, offset)  # 42 (37)
    chemistry["firmware"] = {
        "name": "IntelliChem Firmware Version",
        "value": f"{vMajor}.{vMinor:03}",
    }

    chemistry[f"unknown_at_offset_{offset:02}"], offset = getSome("B", buff, offset)
    chemistry[f"unknown_at_offset_{offset:02}"], offset = getSome("B", buff, offset)
    chemistry[f"unknown_at_offset_{offset:02}"], offset = getSome("B", buff, offset)
    chemistry[f"unknown_at_offset_{offset:02}"], offset = getSome("B", buff, offset)


async def async_request_set_chem_data(
    protocol: ScreenLogicProtocol,
    ph_setpoint: int,
    orp_setpoint: int,
    calcium: int,
    alkalinity: int,
    cyanuric: int,
    salt: int,
):
    return (
        await async_make_request(
            protocol,
            CODE.SETCHEMDATA_QUERY,
            struct.pack(
                "<7I",
                0,
                ph_setpoint,
                orp_setpoint,
                calcium,
                alkalinity,
                cyanuric,
                salt,
            ),
        )
        == b""
    )



import asyncio
import pytest

from .const_data import FAKE_CONFIG_RESPONSE_LARGE
from screenlogicpy.const import CODE as MSG_CODE, MESSAGE as CONST_MESSAGE
from screenlogicpy.requests.protocol import ScreenLogicProtocol
from screenlogicpy.requests.utility import makeMessage


@pytest.mark.asyncio
async def test_async_data_received(event_loop):
    CODE = 1196
    MESSAGE = b"success"

    data = {}

    callback_triggered: asyncio.Future
    callback_triggered = event_loop.create_future()

    async def callback(message, target_data):
        target_data["result"] = message
        callback_triggered.set_result(True)

    protocol = ScreenLogicProtocol(event_loop)
    protocol.register_async_message_callback(CODE, callback, data)

    payload = makeMessage(0, CODE, MESSAGE)
    protocol.data_received(payload)

    await callback_triggered

    assert data.get("result") == MESSAGE


@pytest.mark.asyncio
async def test_async_large_data_received(event_loop):
    CODE = MSG_CODE.CTRLCONFIG_QUERY + 1
    MESSAGE = FAKE_CONFIG_RESPONSE_LARGE[CONST_MESSAGE.HEADER_LENGTH :]

    data = {}

    callback_triggered: asyncio.Future
    callback_triggered = event_loop.create_future()

    async def callback(message, target_data):
        target_data["result"] = message
        callback_triggered.set_result(True)

    protocol = ScreenLogicProtocol(event_loop)
    protocol.register_async_message_callback(CODE, callback, data)

    payload = makeMessage(0, CODE, MESSAGE)
    protocol.data_received(payload[:1024])
    protocol.data_received(payload[1024:])

    await callback_triggered

    assert data.get("result") == MESSAGE


@pytest.mark.asyncio
async def test_async_disconnect(event_loop):
    fut_manager = ScreenLogicProtocol.FutureManager(event_loop)

    futures = []
    for i in range(5):
        futures.append(fut_manager.create(i))

    for x in range(4):
        futures[x].set_result(True)
        assert not fut_manager.all_done()
    futures[4].set_result(True)
    assert fut_manager.all_done()
